#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ads
from ads.secrets import SecretKeeper, Secret
import json


logger = ads.getLogger("ads.secrets")

from dataclasses import dataclass, field


@dataclass
class OracleDBSecret(Secret):
    """
    Dataclass representing the attributes managed and serialized by OracleDBSecretKeeper
    """

    user_name: str
    password: str
    host: str
    port: str
    service_name: str = field(default=None)
    sid: str = field(default=None)
    dsn: str = field(default=None)


class OracleDBSecretKeeper(SecretKeeper):
    """
    `OracleDBSecretKeeper` provides an interface to save Oracle database credentials.
    If you use Wallet file for connnecting to the database, please use ``ADBSecretKeeper``.


    Examples
    --------
    >>> from ads.secrets.oracledb import OracleDBSecretKeeper
    >>> vault_id = "ocid1.vault.oc1..<unique_ID>"
    >>> key_id = "ocid1.key..<unique_ID>"

    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> connection_parameters={
    ...     "user_name":"<your user name>",
    ...     "password":"<your password>",
    ...     "service_name":"service_name",
    ...     "host":"<db host>",
    ...     "port":"<db port>",
    ... }
    >>> oracledb_keeper = OracleDBSecretKeeper(vault_id=vault_id, key_id=key_id, **connection_parameters)
    >>> oracledb_keeper.save("oracledb_employee", "My DB credentials", freeform_tags={"schema":"emp"})
    >>> print(oracledb_keeper.secret_id) # Prints the secret_id of the stored credentials
    >>> oracledb_keeper.export_vault_details("oracledb_employee_att.json") # Save the secret id and vault info to a json file

    >>> # Loading credentails
    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> from ads.secrets.oracledb import OracleDBSecretKeeper
    >>> secret_id = "ocid1.vaultsecret.oc1..<unique_ID>"
    >>> with OracleDBSecretKeeper.load_secret(source=secret_id) as oracledb_creds:
    ...     import pandas as pd
    ...     df = pd.DataFrame.ads.read_sql("select * from EMPLOYEE", connection_parameters=oracledb_creds)


    >>> myoracledb_creds = OracleDBSecretKeeper.load_secret(source='oracledb_employee_att.json', format="json")
    >>> pd.DataFrame.ads.read_sql("select * from ATTRITION_DATA", connection_parameters=myoracledb_creds.to_dict()).head(2)

    """

    def __init__(
        self,
        user_name: str = None,
        password: str = None,
        service_name: str = None,
        sid: str = None,
        host: str = None,
        port: str = "1521",
        dsn: str = None,
        repository_path: str = None,
        repository_key: str = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        user_name: (str, optional). Default None
            user_name of the database
        password: (str, optional). Default None
            password  for connecting to the database
        service_name: (str, optional). Default None
            service name of the Oracle DB instance
        sid: (str, optional). Default None
            Provide sid if service name is not available.
        host: (str, optional). Default None
            Database host name
        port: (str, optional). Default 1521
            Port number
        dsn: (str, optional). Default None
            `dsn` string for connecting with oracledb. Refer `cx_Oracle` documentation
        repository_path: (str, optional). Default None.
            Path to credentials repository. For more details refer `ads.database.connection`
        repository_key: (str, optional). Default None.
            Configuration key for loading the right configuration from repository. For more details refer `ads.database.connection`
        kwargs:
            vault_id: str. OCID of the vault where the secret is stored. Required for saving secret.
            key_id: str. OCID of the key used for encrypting the secret. Required for saving secret.
            compartment_id: str. OCID of the compartment where the vault is located. Required for saving secret.
            auth: dict. Dictionay returned from ads.common.auth.api_keys() or ads.common.auth.resource_principal(). By default, will follow what is set in `ads.set_auth`.  Use this attribute to override the default.
        """
        if repository_path and repository_key:
            from ads.database import connection

            config_from_repo = connection.get_repository(
                repository_key, repository_path
            )
            if user_name is None:
                user_name = config_from_repo.get("username") or config_from_repo.get(
                    "user_name"
                )
            if password is None:
                password = config_from_repo.get("password")
            if service_name is None:
                service_name = config_from_repo.get("service_name")
            if host is None:
                host = config_from_repo.get("host")
            if port is None:
                port = config_from_repo.get("port")
            if sid is None:
                sid = config_from_repo.get("sid")

        self.data = OracleDBSecret(
            user_name=user_name,
            password=password,
            service_name=service_name,
            sid=sid,
            host=host,
            port=port,
            dsn=dsn,
        )

        super().__init__(**kwargs)

    def decode(self) -> "ads.secrets.oracledb.OracleDBSecretKeeper":
        """
        Converts the content in `self.encoded` to `OracleDBSecret` and stores in `self.data`

        Returns
        -------
        OracleDBSecretKeeper:
            Returns self object
        """
        content = json.loads(self._decode())
        logger.debug(f"Decoded secret contains following keys - {list(content.keys())}")
        self.data = OracleDBSecret(
            **content,
        )
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ast
import base64
import json
import oci.vault
import uuid

from oci.vault import VaultsClientCompositeOperations
from oci.vault.models import Base64SecretContentDetails
from oci.vault.models import CreateSecretDetails
from oci.vault.models import UpdateSecretDetails

import ads
from ads.common import oci_client, auth as authutil
from ads.config import NB_SESSION_COMPARTMENT_OCID

logger = ads.getLogger("ads.vault")


class Vault:
    def __init__(
        self,
        vault_id: str = None,
        key_id: str = None,
        compartment_id=None,
        secret_client_auth=None,
        vault_client_auth=None,
        auth=None,
    ):

        """
        Parameters
        ----------
        vault_id: (str, optional). Default None
            ocid of the vault
        key_id: (str, optional). Default None
            ocid of the key that is used for encrypting the content
        compartment_id: (str, optional). Default None
            ocid of the compartment_id where the vault resides. When available in
            environment variable - `NB_SESSION_COMPARTMENT_OCID`, will defult to that.
        secret_client_auth: (dict, optional, deprecated since 2.5.1). Default None.
            deprecated since 2.5.1. Use `auth` instead
        vault_client_auth: (dict, optional, deprecated since 2.5.1). Default None.
            deprecated since 2.5.1. Use `auth` instead
        auth: (dict, optional)
            Dictionay returned from ads.common.auth.api_keys() or ads.common.auth.resource_principal(). By default, will follow what is set in `ads.set_auth`.  Use this attribute to override the default.

        """
        # vault_client_auth. User should use auth to setup the authentication

        self.id = vault_id
        self.key_id = key_id
        self.client_auth = auth or authutil.default_signer()
        if secret_client_auth:
            logger.warning(
                "Deprecated use of `secret_client_auth` since version 2.5.1. Use `auth` instead."
            )
        if vault_client_auth:
            logger.warning(
                "Deprecated use of `vault_client_auth` since version 2.5.1. Use `auth` instead."
            )
        self.secret_client_auth = secret_client_auth or self.client_auth
        self.vault_client_auth = vault_client_auth or self.client_auth

        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )

        self.secret_client = oci_client.OCIClientFactory(
            **self.secret_client_auth
        ).secret
        self.vaults_client_composite = VaultsClientCompositeOperations(
            oci_client.OCIClientFactory(**self.vault_client_auth).vault
        )

    def create_secret(
        self,
        value: dict,
        secret_name: str = None,
        description: str = None,
        encode=True,
        freeform_tags: dict = None,
        defined_tags: dict = None,
    ) -> str:
        """
        Saves value into vault as a secret.

        Parameters
        ----------
        value: dict
            The value to store as a secret.
        secret_name: str, optional
            The name of the secret.
        description: str, optional
            The description of the secret.
        encode: (bool, optional). Default True
            Whether to encode using the default encoding.
        freeform_tags: (dict, optional). Default None
            freeform_tags as defined by the oci sdk
        defined_tags: (dict, optional). Default None
            defined_tags as defined by the oci sdk

        Returns
        -------
        The secret ocid that correspond to the value saved as a secret into vault.
        """
        if not isinstance(self.compartment_id, str):
            raise ValueError("compartment_id needs to be a string.")

        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")

        if self.id is None:
            raise ValueError("vault_id needs to be specified in the constructor.")
        if self.key_id is None:
            raise ValueError("key_id needs to be specified in the constructor.")
        # Encode the secret.
        secret_content_details = self._encode_secret(value, encode=encode)

        # Bundle the secret and metadata about it.
        secrets_details = CreateSecretDetails(
            compartment_id=self.compartment_id,
            description=description
            if description is not None
            else "Data Science service secret",
            secret_content=secret_content_details,
            secret_name=secret_name
            if secret_name is not None
            else "DataScienceSecret_{}".format(str(uuid.uuid4())[-6:]),
            vault_id=self.id,
            key_id=self.key_id,
            freeform_tags=freeform_tags,
            defined_tags=defined_tags,
        )

        # Store secret and wait for the secret to become active.
        secret = self.vaults_client_composite.create_secret_and_wait_for_state(
            create_secret_details=secrets_details,
            wait_for_states=[oci.vault.models.Secret.LIFECYCLE_STATE_ACTIVE],
        ).data
        return secret.id

    def update_secret(
        self,
        secret_id: str,
        secret_content: dict,
        encode: bool = True,
    ) -> str:
        """
        Updates content of a secret.

        Parameters
        ----------
        secret_id: str
            The secret id where the stored secret will be updated.
        secret_content: dict,
            The updated content.
        encode: (bool, optional). Default True
            Whether to encode the secret_content using default encoding

        Returns
        -------
        The secret ocid with updated content.
        """

        if not isinstance(self.compartment_id, str):
            raise ValueError("compartment_id needs to be a string.")

        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")

        if self.id is None:
            raise ValueError("vault_id needs to be specified in the constructor.")
        if self.key_id is None:
            raise ValueError("key_id needs to be specified in the constructor.")
        # Encode the secret.
        secret_content_details = (
            self._encode_secret(secret_content) if encode else secret_content
        )

        # Store the details to update.
        secrets_details = UpdateSecretDetails(
            secret_content=secret_content_details,
        )

        # Create new secret version and wait for the new version to become active.
        secret_update = self.vaults_client_composite.update_secret_and_wait_for_state(
            secret_id,
            secrets_details,
            wait_for_states=[oci.vault.models.Secret.LIFECYCLE_STATE_ACTIVE],
        ).data

        return secret_update.id

    def get_secret(self, secret_id: str, decoded=True) -> dict:
        """
        Retrieve secret content based on the secret ocid provided

        Parameters
        ----------
        secret_id: str
            The secret ocid.
        decoded: (bool, optional). Default True
            Whether to decode the content that is retrieved from vault service using the default decoder.
        Returns
        -------
        The secret content as a dictionary.
        """
        secret_bundle = self.secret_client.get_secret_bundle(secret_id)
        if decoded:
            secret_content = self._secret_to_dict(
                secret_bundle.data.secret_bundle_content.content
            )
            return ast.literal_eval(secret_content)
        else:
            return secret_bundle.data.secret_bundle_content.content

    def _encode_secret(self, secret_content, encode=True):
        secret_content_details = Base64SecretContentDetails(
            content_type=oci.vault.models.SecretContentDetails.CONTENT_TYPE_BASE64,
            stage=oci.vault.models.SecretContentDetails.STAGE_CURRENT,
            content=self._dict_to_secret(secret_content) if encode else secret_content,
        )
        return secret_content_details

    @staticmethod
    def _dict_to_secret(values):
        return base64.b64encode(json.dumps(values).encode("ascii")).decode("ascii")

    @staticmethod
    def _secret_to_dict(secret_content):
        return base64.b64decode(secret_content.encode("ascii")).decode("ascii")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import functools
import os


class NotSupportedError(Exception):
    pass


def experimental(cls):
    @functools.wraps(cls)
    def wrapper(*args, **kwargs):
        instance = cls(*args, **kwargs)
        print(f"{cls.__name__} is experimental and may be removed in the future.")
        return instance

    return wrapper


class PY4JGateway:
    def __init__(self) -> None:
        try:
            from py4j.java_gateway import GatewayParameters, JavaGateway, launch_gateway
        except ModuleNotFoundError:
            raise ModuleNotFoundError("py4j is not installed.")
        if "CONDA_PREFIX" not in os.environ or not os.path.exists(
            os.path.join(os.environ.get("CONDA_PREFIX"), "text-extraction-tools.jar")
        ):
            raise NotSupportedError(
                "Tika is not supported in this distribution. Use alternatives such as pdfplumber."
            )
        port = launch_gateway(
            java_path="/usr/bin/java",
            classpath=os.path.join(
                os.environ.get("CONDA_PREFIX"), "text-extraction-tools.jar"
            ),
        )
        self.gateway = JavaGateway(gateway_parameters=GatewayParameters(port=port))

    def __enter__(self) -> None:
        return self.gateway

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.gateway.shutdown()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np
from sklearn import preprocessing

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import utils
from ads.type_discovery.typed_feature import CategoricalTypedFeature
from ads.type_discovery.typed_feature import ContinuousTypedFeature
from ads.type_discovery.typed_feature import OrdinalTypedFeature

ABSOLUTE_SKEWNESS_THRESHOLD = 0.5
# between 0 and 1, ratio of least / most represented class
CLASS_IMBALANCE_THRESHOLD = 0.5


class TargetVariable:
    """
    This class provides target specific APIs.
    """

    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def __init__(self, sampled_ds, target, target_type):
        from scipy.stats import skew

        self.sampled_ds = sampled_ds
        self.name = target
        self.type = target_type
        self.target_vals = None
        if isinstance(self.type, CategoricalTypedFeature):
            self.target_vals = self.sampled_ds.sampled_df[target].unique().tolist()
            val_counts = self.sampled_ds.sampled_df[target].value_counts()
            self.class_imbalance_ratio = val_counts.min() / val_counts.max()
        elif isinstance(self.type, ContinuousTypedFeature):
            try:
                self.skewness = np.abs(self.sampled_ds.sampled_df[target].skew())
            except TypeError as e:
                self.skewness = np.abs(
                    self.sampled_ds.sampled_df[target].astype("float").skew()
                )
        else:
            # can also be DateTimeTypedFeature IPAddressTypedFeature PhoneNumberTypedFeature GISTypedFeature
            # AddressTypedFeature DocumentTypedFeature ZipcodeTypedFeature UnknownTypedFeature ConstantTypedFeature
            # CreditCardTypedFeature OrdinalTypedFeature
            self.skewness = None
        self.numeric_columns = (
            self.sampled_ds.sampled_df._get_numeric_data().columns.values
        )

    def show_in_notebook(self, feature_names=None):  # pragma: no cover
        """
        Plot target distribution or target versus feature relation.

        Parameters
        ----------
        feature_names: list, Optional
            Plot target against a list of features.
            Display target distribution if feature_names is not provided.
        """
        if not utils.is_notebook():
            print("show_in_notebook called but not in notebook environment")
            return

        verbose = True
        if feature_names is not None:
            for feature_name in feature_names:
                self.sampled_ds.plot(
                    feature_name, self.name, verbose=verbose
                ).show_in_notebook()
                verbose = False
        else:
            self.sampled_ds.plot(self.name, verbose=False).show_in_notebook()

    def is_balanced(
        self,
        skewness_threshold=ABSOLUTE_SKEWNESS_THRESHOLD,
        class_imbalance_threshold=CLASS_IMBALANCE_THRESHOLD,
    ):
        """
        Returns True if the target is balanced, False otherwise.

        Returns
        -------
        is_balanced: bool
        """
        if isinstance(self.type, ContinuousTypedFeature):
            return self.skewness < skewness_threshold
        elif isinstance(self.type, CategoricalTypedFeature):
            return self.class_imbalance_ratio > class_imbalance_threshold
        else:
            return True



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""Contains Mixins for integrating OCI data models
"""
import inspect
import json
import logging
import os
import re
import traceback
from datetime import date, datetime
from typing import Callable, Optional, Union
from enum import Enum

import oci
import yaml
from ads.common import auth
from ads.common.decorator.utils import class_or_instance_method
from ads.common.utils import camel_to_snake
from ads.config import COMPARTMENT_OCID
from dateutil import tz
from dateutil.parser import parse
from oci._vendor import six

logger = logging.getLogger(__name__)

LIFECYCLE_STOP_STATE = ("SUCCEEDED", "FAILED", "CANCELED", "DELETED")


class MergeStrategy(Enum):
    OVERRIDE = "override"
    MERGE = "merge"


class OCIModelNotExists(Exception):
    pass


class OCIClientMixin:
    """Mixin class for representing OCI resource/service with OCI client.

    Most OCI requests requires a client for making the connection.
    Usually the same client will be used for the requests related the same resource/service type.
    This Mixin adds a "client" property to simplify accessing the client.
    The actual client will be initialize lazily so that it is not required for a sub-class
    To use the client, sub-class should override the init_client() method pass in the "client" keyword argument.
    For example:

    @class_or_instance_method
    def init_client(cls, **kwargs) -> oci.logging.LoggingManagementClient:
        return cls._init_client(client=oci.logging.LoggingManagementClient, **kwargs)

    Instance methods in the sub-class can use self.client to access the client.
    The init_client() method is a class method used to create the client.
    Any class method using the client should use init_client() to create the client.
    The call to this method may come from an instance or a class.
    When the method is called from a class,
        the default authentication configured at ADS level with ads.set_auth() will be used.
    When the method is called from an instance,
        the config, signer and kwargs specified when initializing the instance will be used.
    The sub-class's __init__ method should take config, signer and client_kwargs as argument,
        then pass them to the __init__ method of this class.
    This allows users to override the authentication and client initialization parameters.
    For example, log_group = OCILogGroup(config=config, signer=signer, client_kwargs=kwargs)
    """

    config = None
    signer = None
    kwargs = None

    @class_or_instance_method
    def _get_auth(cls):
        client_kwargs = dict(retry_strategy=oci.retry.DEFAULT_RETRY_STRATEGY)
        if cls.kwargs:
            client_kwargs.update(cls.kwargs)

        if cls.config is None and cls.signer is None:
            oci_auth = auth.default_signer(client_kwargs)
        elif not cls.signer and cls.config:
            oci_auth = {"config": cls.config, "client_kwargs": client_kwargs}
        else:
            oci_auth = {
                "config": cls.config,
                "signer": cls.signer,
                "client_kwargs": client_kwargs,
            }
        return oci_auth

    @class_or_instance_method
    def init_client(cls, **kwargs):
        """Initializes the OCI client specified in the "client" keyword argument
        Sub-class should override this method and call cls._init_client(client=OCI_CLIENT)

        Parameters
        ----------
        **kwargs :
            Additional keyword arguments for initializing the OCI client.

        Returns
        -------
            An instance of OCI client.

        """
        return cls._init_client(**kwargs)

    @class_or_instance_method
    def _init_client(cls, client, **kwargs):
        """Initializes the OCI client specified in the "client" argument

        Parameters
        ----------
        client :
            The OCI client class to be initialized, e.g., oci.data_science.DataScienceClient
        **kwargs :
            Additional keyword arguments for initializing the OCI client.

        Returns
        -------
            An instance of OCI client.

        """
        auth = cls._get_auth()
        auth_kwargs = auth.pop("client_kwargs", {})
        if kwargs:
            auth_kwargs.update(kwargs)
        auth.update(auth_kwargs)
        return client(**auth)

    @class_or_instance_method
    def create_instance(obj, *args, **kwargs):
        """Creates an instance using the same authentication as the class or an existing instance.
        If this method is called by a class, the default ADS authentication method will be used.
        If this method is called by an instance, the authentication method set in the instance will be used.
        """
        # Here cls could be a class or an instance.
        # If it is a class, it can be used to create the new instance directly.
        # If it is an instance, the new instance should be created using cls.__class__.
        # Calling the classmethod _create_instance() will make sure that class will be used.
        return obj._create_instance(
            config=obj.config,
            signer=obj.signer,
            client_kwargs=obj.kwargs,
            *args,
            **kwargs,
        )

    @classmethod
    def _create_instance(cls, *args, **kwargs):
        """Initialize an instance."""
        return cls(*args, **kwargs)

    def __init__(self, config=None, signer=None, client_kwargs=None) -> None:
        """Initializes a service/resource with OCI client as a property.
        If config or signer is specified, it will be used to initialize the OCI client.
        If neither of them is specified, the client will be initialized with ads.common.auth.default_signer.
        If both of them are specified, both of them will be passed into the OCI client,
            and the authentication will be determined by OCI Python SDK.

        Parameters
        ----------
        config : dict, optional
            OCI API key config dictionary, by default None.
        signer : oci.signer.Signer, optional
            OCI authentication signer, by default None.
        client_kwargs : dict, optional
            Additional keyword arguments for initializing the OCI client.
        """
        super().__init__()
        self.config = config
        self.signer = signer
        self.kwargs = client_kwargs
        self._client = None

    @property
    def auth(self) -> dict:
        """The ADS authentication config used to initialize the client.
        This auth has the same format as those obtained by calling functions in ads.common.auth.
        The config is a dict containing the following key-value pairs:
        config: The config contains the config loaded from the configuration loaded from `oci_config`.
        signer: The signer contains the signer object created from the api keys.
        client_kwargs: client_kwargs contains the `client_kwargs` that was passed in as input parameter.
        """
        return self._get_auth()

    @property
    def client(self):
        """OCI client"""
        if self._client is None:
            self._client = self.init_client()
        return self._client


class OCISerializableMixin(OCIClientMixin):
    """Mixin class containing OCI serialization/de-serialization methods.
    These methods are copied and modified from the OCI BaseClient.

    """

    type_mappings = None

    def serialize(self):
        """Serialize the model to a dictionary that is ready to be send to OCI API.

        Returns
        -------
        dict
            A dictionary that is ready to be send to OCI API.

        """
        return self.client.base_client.sanitize_for_serialization(self)

    @staticmethod
    def _parse_kwargs(attribute_map: dict, **kwargs):
        """Parse kwargs to make all key in camel format."""
        parsed_kwargs = {}
        for key, val in kwargs.items():
            if key in attribute_map:
                parsed_kwargs[attribute_map[key]] = val
            else:
                parsed_kwargs[key] = val

        return parsed_kwargs

    @classmethod
    def deserialize(cls, data, to_cls):
        """De-serialize data from dictionary to an OCI model"""
        if cls.type_mappings is None:
            cls.type_mappings = cls.init_client().base_client.type_mappings

        if data is None:
            return None

        # This is a work around for enums not being present
        # in the type mappings.
        # See OraclePythonSdkCodegen removeEnumsFromModelGeneration().
        if to_cls in cls.type_mappings:
            to_cls = cls.type_mappings[to_cls]

            if isinstance(data, dict):
                data = cls._parse_kwargs(to_cls().attribute_map, **data)
        else:
            return cls.__deserialize_primitive(data, to_cls)

        if hasattr(to_cls, "get_subtype"):
            # Return the object as is if it is already deserialized.
            if isinstance(data, to_cls) or issubclass(data.__class__, to_cls):
                return data

            # Use the discriminator value to get the correct subtype.
            to_cls = to_cls.get_subtype(data)  # get_subtype returns a str
            to_cls = cls.type_mappings[to_cls]

            # kwargs needs to be parsed again as there are more attributes in the sub types.
            if isinstance(data, dict):
                data = cls._parse_kwargs(to_cls().attribute_map, **data)

        if to_cls in [int, float, six.u, bool]:
            return cls.__deserialize_primitive(data, to_cls)
        elif to_cls == object:
            return data
        elif to_cls == date:
            return cls.__deserialize_date(data)
        elif to_cls == datetime:
            return cls.__deserialize_datetime(data)
        else:
            return cls.__deserialize_model(data, to_cls)

    @classmethod
    def __deserialize_model(cls, data, to_cls):
        """De-serializes list or dict to model."""
        if isinstance(data, to_cls):
            return data

        instance = to_cls()

        for attr, attr_type in instance.swagger_types.items():
            property = instance.attribute_map[attr]
            if property in data:
                value = data[property]
                setattr(instance, attr, cls.deserialize(value, attr_type))

        return instance

    @staticmethod
    def __deserialize_primitive(data, cls):
        """De-serializes string to primitive type."""
        try:
            value = cls(data)
        except UnicodeEncodeError:
            value = data
        except TypeError:
            value = data
        return value

    @staticmethod
    def __deserialize_date(string):
        """De-serializes string to date."""
        try:
            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise Exception("Failed to parse `{0}` into a date object".format(string))

    @staticmethod
    def __deserialize_datetime(string):
        """De-serializes string to datetime.

        The string should be in iso8601 datetime format.
        """
        if isinstance(string, datetime):
            return string
        try:
            # If this parser creates a date without raising an exception
            # then the time zone is utc and needs to be set.
            naivedatetime = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%fZ")
            awaredatetime = naivedatetime.replace(tzinfo=tz.tzutc())
            return awaredatetime

        except ValueError:
            try:
                return parse(string)
            except ImportError:
                return string
            except ValueError:
                raise Exception(
                    "Failed to parse `{0}` into a datetime object".format(string)
                )
        except ImportError:
            return string


class OCIModelMixin(OCISerializableMixin):
    """Mixin class to operate OCI model.
    OCI resources are represented by models in the OCI Python SDK.

    Unifying OCI models for the same resource
    -----------------------------------------
    OCI SDK uses different models to represent the same resource for different operations.
    For example, CreateLogDetails is used when creating a log resource,
        while LogSummary is returned when listing the log resources.
    However, both CreateLogDetails and LogSummary have the same commonly used attribute like
        compartment_id, display_name, log_type, etc.
    In general, there is a class with a super set of all properties.
    For example, the Log class contains all properties of CreateLogDetails and LogSummary,
        as well as other classes representing an OCI log resource.
    A subclass can be implemented with this Mixin to unify the OCI models,
        so that all properties are available to the user.
    For example, if we define the Mixin model as ``class OCILog(OCIModelMixin, oci.logging.models.Log)``,
        users will be able to access properties like ``OCILog().display_name``
    Since this sub-class contains all the properties, it can be converted to any related OCI model in an operation.
    For example, we can create ``CreateLogDetails`` from ``OCILog`` by extracting a subset of the properties.
    When listing the resources, properties from ``LogSummary`` can be used to update
        the corresponding properties of ``OCILog``.
    Such convertion can be done be the generic methods provided by this Mixin.
    Although OCI SDK accepts dictionary (JSON) data instead of objects like CreateLogDetails when creating or
        updating the resource, the structure and keys of the dictionary is not easy for a user to remember.
    It is also unnecessary for the users to construct the entire dictionary if they only want to update a single value.

    This Mixin class should be the first parent as the class from OCI SDK does not call ``super().__init__()``
        in its ``__init__()`` constructor.
    Mixin properties may not be intialized correctly if ``super().__init__()`` is not called.


    Provide generic methods for CRUDL operations
    --------------------------------------------
    Since OCI SDK uses different models in CRUDL operations,
        this Mixin provides the following method to convert between them.
    An OCI model instance can be any OCI model of the resource containing some properties, e.g. LogSummary
    ``from_oci_model()`` static method can be used to initialize a new instance from an OCI model instance.
    ``update_from_oci_model()`` can be used to update the existing properties from an OCI model instance.
    ``to_oci_model()`` can be used to extract properties from the Mixin model to OCI model.

    """

    # Regex pattern matching the module name of an OCI model.
    OCI_MODEL_PATTERN = r"oci.[^.]+\.models[\..*]?"
    # Constants
    CONS_COMPARTMENT_ID = "compartment_id"

    @staticmethod
    def check_compartment_id(compartment_id: Optional[str]) -> str:
        """Checks if a compartment ID has value and
            return the value from NB_SESSION_COMPARTMENT_OCID environment variable if it is not specified.

        Parameters
        ----------
        compartment_id : str
            Compartment OCID or None

        Returns
        -------
        type
            str: Compartment OCID

        Raises
        ------
        ValueError
            compartment_id is not specified and NB_SESSION_COMPARTMENT_OCID environment variable is not set

        """
        compartment_id = compartment_id or COMPARTMENT_OCID
        if not compartment_id:
            raise ValueError("Specify compartment OCID.")
        return compartment_id

    @class_or_instance_method
    def list_resource(
        cls, compartment_id: str = None, limit: int = 0, **kwargs
    ) -> list:
        """Generic method to list OCI resources

        Parameters
        ----------
        compartment_id : str
            Compartment ID of the OCI resources. Defaults to None.
            If compartment_id is not specified,
            the value of NB_SESSION_COMPARTMENT_OCID in environment variable will be used.
        limit : int
            The maximum number of items to return. Defaults to 0, All items will be returned
        **kwargs :
            Additional keyword arguments to filter the resource.
            The kwargs are passed into OCI API.

        Returns
        -------
        list
            A list of OCI resources

        Raises
        ------
        NotImplementedError
            List method is not supported or implemented.

        """
        if limit:
            items = cls._find_oci_method("list")(
                cls.check_compartment_id(compartment_id), limit=limit, **kwargs
            ).data
        else:
            items = oci.pagination.list_call_get_all_results(
                cls._find_oci_method("list"),
                cls.check_compartment_id(compartment_id),
                **kwargs,
            ).data
        return [cls.from_oci_model(item) for item in items]

    @classmethod
    def _find_oci_parent(cls):
        """Finds the parent OCI model class

        Parameters
        ----------

        Returns
        -------
        class
            An OCI model class

        Raises
        ------
        AttributeError
            If the class if not inherit from an OCI model.

        """
        oci_model = None
        for parent in inspect.getmro(cls):
            if re.match(OCIModelMixin.OCI_MODEL_PATTERN, parent.__module__):
                oci_model = parent
                break
        if not oci_model:
            raise AttributeError(f"{cls.__name__} is not inherited from an OCI model.")
        return oci_model

    @class_or_instance_method
    def _find_oci_method(cls, method: str) -> Callable:
        """Finds the OCI method for operations like get/list

        Parameters
        ----------
        method : str
            The operation to be performed, e.g. get, list, delete


        Returns
        -------
        callable
            The method from OCI client to perform the operation.

        """
        client = cls.init_client()

        user_define_method = f"oci_{method}_method".upper()
        if hasattr(cls, user_define_method) and getattr(cls, user_define_method):
            return getattr(cls, user_define_method).__get__(client, client.__class__)

        oci_class = cls._find_oci_parent()
        if method == "list":
            method_name = f"{method}_{camel_to_snake(oci_class.__name__)}s"
        else:
            method_name = f"{method}_{camel_to_snake(oci_class.__name__)}"

        if hasattr(client, method_name):
            return getattr(client, method_name)

        raise NotImplementedError(
            f"{str(oci_class)} does not have {method_name} method. "
            f"Define {user_define_method} in {str(cls)}"
        )

    @class_or_instance_method
    def from_oci_model(cls, oci_instance):
        """Initialize an instance from an instance of OCI model.

        Parameters
        ----------
        oci_instance :
            An instance of an OCI model.

        """
        oci_model = cls._find_oci_parent()
        kwargs = {}
        for attr in oci_model().swagger_types.keys():
            if hasattr(oci_instance, attr):
                kwargs[attr] = getattr(oci_instance, attr)
        instance = cls.create_instance(**kwargs)
        if hasattr(oci_instance, "attribute_map"):
            instance._oci_attributes = oci_instance.attribute_map
        return instance

    @class_or_instance_method
    def from_dict(cls, data):
        """Initialize an instance from a dictionary.

        Parameters
        ----------
        data : dict
            A dictionary containing the properties to initialize the class.

        """
        return cls.create_instance(**data)

    @classmethod
    def deserialize(cls, data: dict, to_cls: str = None):
        """Deserialize data

        Parameters
        ----------
        data : dict
            A dictionary containing the data to be deserialized.

        to_cls : str
            The name of the OCI model class to be initialized using the data.
            The OCI model class must be from the same OCI service of the OCI client (self.client).
            Defaults to None, the parent OCI model class name will be used
            if current class is inherited from an OCI model.
            If parent OCI model class is not found or not from the same OCI service,
            the data will be returned as is.

        """
        if to_cls is None:
            oci_model = cls._find_oci_parent()
            to_cls = oci_model.__name__
        return super().deserialize(data, to_cls)

    @class_or_instance_method
    def _get(cls, ocid):
        get_method = cls._find_oci_method("get")
        return get_method(ocid).data

    @class_or_instance_method
    def from_ocid(cls, ocid: str):
        """Initializes an object from OCID

        Parameters
        ----------
        ocid : str
            The OCID of the object

        """
        oci_instance = cls._get(ocid)
        return cls.from_oci_model(oci_instance)

    def __init__(
        self,
        config: dict = None,
        signer: oci.signer.Signer = None,
        client_kwargs: dict = None,
        **kwargs,
    ) -> None:
        # Initialize an empty object
        super().__init__(config=config, signer=signer, client_kwargs=client_kwargs)

        if kwargs:
            self.update_from_oci_model(
                self.deserialize(kwargs), merge_strategy=MergeStrategy.MERGE
            )
        # When from_oci_model is called to initialize the object from an OCI model,
        # _oci_attributes stores the attribute names from the OCI model
        # This is used to determine if an additional get request should be called to get additional attributes.
        # For example, some attributes are not included in the results from OCI list request.
        # When user access such attributes, an additional get request should be made to get the data from OCI.
        # See __getattribute__() for the implementation.
        self._oci_attributes = {}

    @property
    def name(self) -> str:
        """Gets the name of the object."""
        if hasattr(self, "display_name"):
            return getattr(self, "display_name")
        return ""

    @name.setter
    def name(self, value: str):
        """Sets the name of the object.

        Parameters
        ----------
        value : str
            The name of the object.
        """
        setattr(self, "display_name", value)

    def load_properties_from_env(self):
        """Loads properties from the environment"""
        env_var_mapping = {
            "project_id": ["PROJECT_OCID"],
            OCIModelMixin.CONS_COMPARTMENT_ID: [
                "JOB_RUN_COMPARTMENT_OCID",
                "NB_SESSION_COMPARTMENT_OCID",
            ],
        }
        for attr, env_names in env_var_mapping.items():
            try:
                env_value = next(
                    os.environ.get(env_name)
                    for env_name in env_names
                    if os.environ.get(env_name, None) is not None
                )
                if getattr(self, attr, "") is None:
                    setattr(self, attr, env_value)
            except:
                pass

    def to_oci_model(self, oci_model):
        """Converts the object into an instance of OCI data model.

        Parameters
        ----------
        oci_model : class or str
            The OCI model to be converted to. This can be a string of the model name.
        type_mapping : dict
            A dictionary mapping the models.
            Returns: An instance of the oci_model

        Returns
        -------

        """
        if isinstance(oci_model, str):
            oci_model_name = oci_model
        else:
            oci_model_name = oci_model.__name__
        data = json.dumps(self.to_dict()).encode("utf8")
        return self.client.base_client.deserialize_response_data(data, oci_model_name)

    @staticmethod
    def flatten(data: dict) -> dict:
        """Flattens a nested dictionary.

        Parameters
        ----------
        data : A nested dictionary


        Returns
        -------
        dict
            The flattened dictionary.

        """
        flatten_dict = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flatten_dict.update(OCIModelMixin.flatten(value))
            else:
                flatten_dict[key] = value
        return flatten_dict

    def to_dict(self, flatten: bool = False) -> dict:
        """Converts the properties to a dictionary

        Parameters
        ----------
        flatten :
             (Default value = False)

        Returns
        -------

        """
        data = self.serialize()
        if flatten:
            data = OCIModelMixin.flatten(data)
        return data

    def update_from_oci_model(
        self, oci_model_instance, merge_strategy: MergeStrategy = MergeStrategy.OVERRIDE
    ):
        """Updates the properties from OCI model with the same properties.

        Parameters
        ----------
        oci_model_instance :
            An instance of OCI model, which should have the same properties of this class.
        """
        for attr in self.swagger_types.keys():
            if (
                hasattr(oci_model_instance, attr)
                and getattr(oci_model_instance, attr)
                and (
                    not hasattr(self, attr)
                    or not getattr(self, attr)
                    or merge_strategy == MergeStrategy.OVERRIDE
                )
            ):
                setattr(self, attr, getattr(oci_model_instance, attr))

        if hasattr(oci_model_instance, "attribute_map"):
            self._oci_attributes = oci_model_instance.attribute_map
        return self

    def sync(self, merge_strategy: MergeStrategy = MergeStrategy.OVERRIDE):
        """Refreshes the properties of the object from OCI"""
        return self.update_from_oci_model(
            self.from_ocid(self.id), merge_strategy=merge_strategy
        )

    def delete(self):
        """Deletes the resource"""
        delete_method = self._find_oci_method("delete")
        delete_method(self.id)
        return self

    def __getattribute__(self, name: str):
        """Returns an attribute value of the resource.

        This method will try to sync the values from OCI service when it is not already available locally.
        Some attribute value may not be available locally if the previous OCI API call returns an OCI model
        that contains only a subset of the attributes.
        For example, JobSummary model contains only a subset of the attributes from the Job model.

        Parameters
        ----------
        name : str
            Attribute name.
        """
        skip_lookup = ["id", "attribute_map", "_oci_attributes"]

        if name in skip_lookup or name.startswith("_"):
            return super().__getattribute__(name)

        # Ignore if _oci_attributes is not initialized
        if not hasattr(self, "_oci_attributes"):
            return super().__getattribute__(name)

        if (
            hasattr(self, "attribute_map")
            and name in self.attribute_map
            and name not in self._oci_attributes
            and hasattr(self, "id")
            and self.id
        ):
            # Do not sync if it is in the sync process
            stack = traceback.extract_stack()
            for frame in reversed(stack):
                if frame.name == "sync":
                    return super().__getattribute__(name)
            # Sync only if there is no value
            if not super().__getattribute__(name):
                try:
                    self.sync(merge_strategy=MergeStrategy.MERGE)
                except oci.exceptions.ServiceError as ex:
                    # 400 errors are usually cause by the user
                    if ex.status == 400:
                        logger.error("%s - %s: %s", self.__class__, ex.code, ex.message)
                        self._oci_attributes = self.attribute_map
                    else:
                        logger.error(
                            "Failed to synchronize the properties of %s due to service error:\n%s",
                            self.__class__,
                            str(ex),
                        )
                except Exception as ex:
                    logger.error(
                        "Failed to synchronize the properties of %s: %s\n%s",
                        self.__class__,
                        type(ex),
                        str(ex),
                    )
        return super().__getattribute__(name)

    @property
    def status(self) -> Optional[str]:
        """Status of the object.

        Returns
        -------
        str
            Status of the object.
        """
        if not self.lifecycle_state or not self.lifecycle_state in LIFECYCLE_STOP_STATE:
            self.sync()
        return self.lifecycle_state

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())


class OCIWorkRequestMixin:
    """Mixin class containing methods related to OCI work request"""

    def wait_for_work_request(
        self,
        work_request_id: str,
        wait_for_state: Union[str, tuple],
        max_wait_seconds: int = None,
        wait_interval_seconds: int = None,
    ):
        """Wait for a work request to be completed.

        Parameters
        ----------
        work_request_id : str
            OCI work request ID
        wait_for_state : str or tuple
            The state to wait for. Must be a tuple for multiple states.
        max_wait_seconds : int
            Max wait seconds for the work request. Defaults to None (Default value from OCI SDK will be used).
        wait_interval_seconds : int
            Interval in seconds between each status check. Defaults to None (Default value from OCI SDK will be used).

        Returns
        -------
        Response
            OCI API Response

        """
        result = None
        if hasattr(self.client, "get_work_request") and callable(
            getattr(self.client, "get_work_request")
        ):
            try:
                wait_period_kwargs = {}
                if max_wait_seconds is not None:
                    wait_period_kwargs["max_wait_seconds"] = max_wait_seconds
                if wait_interval_seconds is not None:
                    wait_period_kwargs["max_interval_seconds"] = wait_interval_seconds

                logger.debug(
                    "Action completed. Waiting until the work request has entered state: %s",
                    wait_for_state,
                )
                result = oci.wait_until(
                    self.client,
                    self.client.get_work_request(work_request_id),
                    "status",
                    wait_for_state,
                    **wait_period_kwargs,
                )
            except oci.exceptions.MaximumWaitTimeExceeded:
                # If we fail, we should show an error, but we should still provide the information to the customer
                logger.error(
                    "Failed to wait until the work request %s entered the specified state. Maximum wait time exceeded",
                    work_request_id,
                )
            except Exception:
                logger.error(
                    "Encountered error while waiting for work request to enter the specified state."
                )
                raise
        else:
            logger.error(
                "%s does not support wait for the work request to enter the specified state",
                str(self.client.__class__),
            )
        return result

    def get_work_request_response(
        self,
        response: str,
        wait_for_state: Union[str, tuple],
        success_state: str,
        max_wait_seconds: int = None,
        wait_interval_seconds: int = None,
        error_msg: str = "",
    ):
        if "opc-work-request-id" in response.headers:
            opc_work_request_id = response.headers.get("opc-work-request-id")
            work_request_response = self.wait_for_work_request(
                opc_work_request_id,
                wait_for_state=wait_for_state,
                max_wait_seconds=max_wait_seconds,
                wait_interval_seconds=wait_interval_seconds,
            )
            # Raise an error if the failed to create the resource.
            if work_request_response.data.status != success_state:
                raise oci.exceptions.RequestException(
                    f"{error_msg}\n" + str(work_request_response.data)
                )

        else:
            # This will likely never happen as OCI SDK will raise an error if the HTTP request is not successful.
            raise oci.exceptions.RequestException(
                f"opc-work-request-id not found in response headers: {response.headers}"
            )
        return work_request_response


class OCIModelWithNameMixin:
    """Mixin class to operate OCI model which contains name property."""

    @classmethod
    def from_name(cls, name: str, compartment_id: Optional[str] = None):
        """Initializes an object from name.

        Parameters
        ----------
        name: str
            The name of the object.
        compartment_id: (str, optional). Defaults to None.
            Compartment OCID of the OCI resources. If `compartment_id` is not specified,
            the value will be taken from environment variables.
        """
        res = cls.list_resource(compartment_id=compartment_id, limit=1, name=name)
        if not res:
            raise OCIModelNotExists()
        return cls.from_oci_model(res[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Discrete feature type.

Classes:
    Discrete
        The Discrete feature type.
"""
import matplotlib.pyplot as plt
import pandas as pd
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _count_unique_missing,
    _set_seaborn_theme,
    SchemeTeal,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class Discrete(FeatureType):
    """
    Type representing discrete values.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows distributions of datasets using box plot.
    """

    description = "Type representing discrete values."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count).

        Examples
        --------
        >>> discrete_numbers = pd.Series([35, 25, 13, 42],
                   name='discrete')
        >>> discrete_numbers.ads.feature_type = ['discrete']
        >>> discrete_numbers.ads.feature_stat()
                    discrete
        count	4
        unique	4

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series provided.
        """
        return _count_unique_missing(x)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows distributions of datasets using box plot.

        Examples
        --------
        >>> discrete_numbers = pd.Series([35, 25, 13, 42],
                   name='discrete')
        >>> discrete_numbers.ads.feature_type = ['discrete']
        >>> discrete_numbers.ads.feature_stat()
            Metric  Value
        0	count	4
        1	unique	4

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Discrete feature type.
        """
        col_name = x.name if x.name else "discrete"
        df = x.to_frame(name=col_name)
        df = df.dropna()
        if len(df.index):
            _set_seaborn_theme()
            ax = seaborn.countplot(y=col_name, data=df, color=SchemeTeal.AREA_DARK)
            ax.set(xlabel="Count")
            return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> discrete_numbers = pd.Series([35, 25, 13, 42],
                   name='discrete')
        >>> discrete_numbers.ads.feature_type = ['discrete']
        >>> discrete_numbers.ads.feature_domain()
        constraints: []
        stats:
            count: 4
            unique: 4
        values: Discrete

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the Discrete feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
from typing import Any, Dict, TypeVar, Type
from ads.jobs.serializer import Serializable


Self = TypeVar("Self", bound="Builder")
"""Special type to represent the current enclosed class.

This type is used by factory class method or when a method returns ``self``.
"""


class Builder(Serializable):

    attribute_map = {}

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        """To initialize the object,
        user can either pass in the specification as a dictionary or through keyword arguments.

        Parameters
        ----------
        spec : dict, optional
            Object specification, by default None
        kwargs: dict
            Specification as keyword arguments.
            If spec contains the same key as the one in kwargs, the value from kwargs will be used.
        """
        super().__init__()
        self._spec = self._load_default_properties()
        self._spec.update(self._standardize_spec(spec))
        self._spec.update(self._standardize_spec(kwargs))

    def _load_default_properties(self):
        """
        Load default properties from environment variables, notebook session, etc.
        Should be implemented in the child classes.

        Returns
        -------
        Dict
            A dictionary of default properties.
        """
        return {}

    def _standardize_spec(self, spec) -> dict:
        if not spec:
            return {}
        snake_to_camel_map = {v: k for k, v in self.attribute_map.items()}
        for key in list(spec.keys()):
            if key not in self.attribute_map and key in snake_to_camel_map:
                spec[snake_to_camel_map[key]] = spec.pop(key)
        return spec

    def set_spec(self: Self, k: str, v: Any) -> Self:
        """Sets a specification property for the object.

        Parameters
        ----------
        k: str
            key, the name of the property.
        v: Any
            value, the value of the property.

        Returns
        -------
        Self
            This method returns self to support chaining methods.
        """
        if v is not None:
            self._spec[k] = v
        else:
            self._spec.pop(k, None)
        return self

    def get_spec(self, key: str, default: Any = None) -> Any:
        """Gets the value of a specification property

        Parameters
        ----------
        key : str
            The name of the property.
        default : Any, optional
            The default value to be used, if the property does not exist, by default None.

        Returns
        -------
        Any
            The value of the property.
        """
        return self._spec.get(key, default)

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.
        Subclass should overwrite this value.
        """
        return "builder"

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML.

        This implementation returns the class name with the first letter coverted to lower case.
        """
        class_name = self.__class__.__name__
        return class_name[0].lower() + class_name[1:] if len(class_name) > 1 else ""

    def to_dict(self) -> dict:
        """Converts the object to dictionary with kind, type and spec as keys."""
        spec = copy.deepcopy(self._spec)
        for key, value in spec.items():
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            spec[key] = value

        return {
            "kind": self.kind,
            "type": self.type,
            "spec": spec,
        }

    @classmethod
    def from_dict(cls: Type[Self], obj_dict: dict) -> Self:
        """Initialize the object from a Python dictionary"""
        return cls(spec=obj_dict.get("spec"))

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The ADS accessor for the Pandas Series.
The accessor will be initialized with the pandas object the user is interacting with.

Examples
--------
    >>> from ads.feature_engineering.accessor.series_accessor import ADSSeriesAccessor
    >>> from ads.feature_engineering.feature_type.string import String
    >>> from ads.feature_engineering.feature_type.ordinal import Ordinal
    >>> from ads.feature_engineering.feature_type.base import Tag
    >>> series = pd.Series(['name1', 'name2', 'name3'])
    >>> series.ads.default_type
    'string'
    >>> series.ads.feature_type
    ['string']
    >>> series.ads.feature_type_description
        Feature Type                         Description
    ----------------------------------------------------
    0         string    Type representing string values.
    >>> series.ads.feature_type = ['string', Ordinal, Tag('abc')]
    >>> series.ads.feature_type
    ['string', 'ordinal', 'abc']
    >>> series1 = series.dropna()
    >>> series1.ads.sync(series)
    >>> series1.ads.feature_type
    ['string', 'ordinal', 'abc']
"""

import inspect
import logging
from typing import List, Union

import pandas as pd
from ads.feature_engineering.accessor.mixin.eda_mixin_series import EDAMixinSeries
from ads.feature_engineering.accessor.mixin.feature_types_mixin import (
    ADSFeatureTypesMixin,
)
from ads.feature_engineering.exceptions import TypeNotFound
from ads.feature_engineering.feature_type.base import FeatureType, Tag
from ads.feature_engineering.feature_type.handler.feature_validator import (
    FeatureValidator,
)
from ads.feature_engineering.feature_type.adsstring.string import ADSString

logger = logging.getLogger(__name__)


class ADSSeriesValidator:
    """Class helper to invoke registerred validator on a series level."""

    def __init__(self, feature_type_list: List[FeatureType], series: pd.Series) -> None:
        """Initializes ADS series validator.

        Parameters
        ----------
        feature_type_list : List[FeatureType]
            The list of feature types.
        series : `pd.Series`
            The pandas series.
        """
        self._feature_type_list = feature_type_list
        self._series = series

    def __getattr__(self, attr):
        """Makes it possible to invoke registered validators as a regular method."""
        for feature_type in self._feature_type_list:
            if hasattr(feature_type.validator, attr):
                feature_type.validator._bind_data(self._series)
                return getattr(feature_type.validator, attr)
        raise AttributeError(attr)


@pd.api.extensions.register_series_accessor("ads")
class ADSSeriesAccessor(ADSFeatureTypesMixin, EDAMixinSeries):
    """ADS accessor for Pandas Series.

    Attributes
    ----------
    name: str
        The name of Series.
    tags: List[str]
        The list of tags for the Series.

    Methods
    -------
    help(self, prop: str = None) -> None
        Provids docstring for affordable methods and properties.
    sync(self, src: Union[pd.DataFrame, pd.Series]) -> None
        Syncs feature types of current series with that from src.

    Attributes
    ----------
    default_type(self) -> str
        Gets the name of default feature type for the series.
    feature_type(self) -> List[str]
        Gets the list of registered feature types for the series.
    feature_type_description(self) -> pd.DataFrame
        Gets the list of registered feature types in a DataFrame format.

    Examples
    --------
    >>> from ads.feature_engineering.accessor.series_accessor import ADSSeriesAccessor
    >>> from ads.feature_engineering.feature_type.string import String
    >>> from ads.feature_engineering.feature_type.ordinal import Ordinal
    >>> from ads.feature_engineering.feature_type.base import Tag
    >>> series = pd.Series(['name1', 'name2', 'name3'])
    >>> series.ads.default_type
    'string'
    >>> series.ads.feature_type
    ['string']
    >>> series.ads.feature_type_description
        Feature Type                         Description
    ----------------------------------------------------
    0         string    Type representing string values.
    >>> series.ads.feature_type = ['string', Ordinal, Tag('abc')]
    >>> series.ads.feature_type
    ['string', 'ordinal', 'abc']
    >>> series1 = series.dropna()
    >>> series1.ads.sync(series)
    >>> series1.ads.feature_type
    ['string', 'ordinal', 'abc']
    """

    def __init__(self, pandas_obj: pd.Series) -> None:
        """Initializes ADS Pandas Series Accessor.

        Parameters
        ----------
        pandas_obj : `pd.Series`
            The pandas series
        """
        self._obj = pandas_obj
        super().__init__()
        self._feature_type = [self._default_type]
        self.tags = []
        self.name = self._obj.name

    @property
    def _default_type(self) -> FeatureType:
        """Gets default feature type for the series.

        Returns
        -------
        FeatureType
            The default feature type for the series.
        """
        return self._feature_type_by_dtype(self._obj.dtype)

    @property
    def default_type(self) -> str:
        """Gets the name of default feature type for the series.

        Returns
        -------
        str
            The name of default feature type.
        """
        return self._default_type.name

    @property
    def feature_type(self) -> List[str]:
        """Gets the list of registered feature types for the series.

        Returns
        -------
        List[str]
            Names of feature types.

        Examples
        --------
        >>> series = pd.Series(['name1'])
        >>> series.ads.feature_type = ['name', 'string', Tag('tag for name')]
        >>> series.ads.feature_type
        ['name', 'string', 'tag for name']
        """
        types = []
        for feature_type in self._feature_type:
            types.append(feature_type.name)
        return types + self.tags

    @property
    def feature_type_description(self) -> pd.DataFrame:
        """Gets the list of registered feature types in a DataFrame format.

        Returns
        -------
        pd.DataFrame
            The DataFrame with feature types for this series.

        Examples
        --------
        >>> series = pd.Series(['name1'])
        >>> series.ads.feature_type = ['name', 'string', Tag('Name tag')]
        >>> series.ads.feature_type_description
                Feature Type                               Description
            ----------------------------------------------------------
            0           name            Type representing name values.
            1         string          Type representing string values.
            2        Name tag                                     Tag.
        """
        feature_types = (
            (feature_type.name, feature_type.description)
            for feature_type in self._feature_type
            if self._is_type_registered(feature_type)
        )
        tags = ((tag, "Tag") for tag in self.tags)
        return pd.DataFrame(
            tuple(feature_types) + tuple(tags), columns=["Feature Type", "Description"]
        )

    @feature_type.setter
    def feature_type(self, feature_types: List[Union[FeatureType, str, Tag]]) -> None:
        """Sets feature types for the series.

        Parameters
        ----------
        feature_types : List[Union[FeatureType, str, Tag]]
            The list of feature types.

        Return
        ------
        None
            Nothing.

        Raises
        ------
        TypeError: If input data has wrong format.

        Examples
        --------
        >>> series = pd.Series(['name1', 'name2', 'name3'])
        >>> series.ads.feature_type = ['name']
        >>> series.feature_type
        ['name', 'string']
        >>> series.ads.feature_type = ['string', 'name']
        >>> series.feature_type
        ['string', 'name']
        >>> series.ads.feature_type = []
        >>> series.feature_type
        ['string']
        """
        if feature_types is None or not isinstance(feature_types, list):
            raise TypeError("Argument must be a list of feature types.")

        self._feature_type = []
        self.tags = []
        for feature_type in feature_types:
            self._add_feature_type(feature_type)

        default_feature_type = self._default_type
        if default_feature_type not in self._feature_type:
            self._add_feature_type(default_feature_type)

    def sync(self, src: Union[pd.DataFrame, pd.Series]) -> None:
        """Syncs feature types of current series with that from src.

        The src could be a dataframe or a series. In either case, only columns
        with matched names are synced.

        Parameters
        ----------
        src: (`pd.DataFrame` | `pd.Series`)
            The source to sync from.

        Returns
        -------
        None
            Nothing.

        Examples
        --------
        >>> series = pd.Series(['name1', 'name2', 'name3', None])
        >>> series.ads.feature_type = ['name']
        >>> series.ads.feature_type
        ['name', string]
        >>> series.dropna().ads.feature_type
        ['string']
        >>> series1 = series.dropna()
        >>> series1.ads.sync(series)
        >>> series1.ads.feature_type
        ['name', 'string']
        """
        if isinstance(src, pd.DataFrame):
            if self._obj.name not in src.columns:
                logger.warning(
                    "The source DataFrame doesn't have a clumn %s.", self._obj.name
                )
                return
            self._sync(src[self._obj.name])
        elif isinstance(src, pd.Series):
            self._sync(src)

    def _sync(self, src: pd.Series) -> None:
        """Copies all feature types from src series to the current."""
        new_feature_type = [ft for ft in src.ads._feature_type]
        new_tags = [Tag(tag) for tag in src.ads.tags]
        self.feature_type = new_feature_type + new_tags

    def _add_feature_type(self, feature_type: Union[FeatureType, str, Tag]) -> None:
        """Adds a feature type to the series.

        Parameters
        ----------
        feature_type : Union[FeatureType, str, Tag]
            The feature type to add.

        Returns
        -------
        None
            Nothing.
        """
        if isinstance(feature_type, Tag):
            if feature_type.name in self.feature_type:
                logger.warning(
                    "The tag '%s' is already added to the series '%s'.",
                    feature_type.name,
                    self.name,
                )
            self.tags.append(feature_type.name)
        else:
            feature_type_cls = self._get_type(feature_type)
            if feature_type_cls.name in self.feature_type:
                logger.warning(
                    "The type '%s' is already added to the series '%s'.",
                    feature_type_cls.name,
                    self.name,
                )
            self._feature_type.append(feature_type_cls)

    def _remove_feature_type(self, feature_type: Union[FeatureType, str, Tag]) -> None:
        """Removes a feature type.

        Parameters
        ----------
        feature_type : Union[FeatureType, str, Tag]
            feature type

        Raises
        ------
        TypeNotFound
            Type does not exist for this series
        """
        tag = self._get_tag(feature_type)
        if tag:
            if tag in self.tags:
                self.tags.remove(tag)
                return
            else:
                raise TypeNotFound(tag)

        feature_type_cls = self._get_type(feature_type)
        if feature_type_cls in self._feature_type:
            self._feature_type.remove(feature_type_cls)
            return

        raise TypeNotFound(feature_type_cls.__name__)

    def _get_tag(self, feature_type: Union[str, Tag]) -> str:
        if isinstance(feature_type, str) and feature_type in self.tags:
            return feature_type
        if isinstance(feature_type, Tag):
            return feature_type.name

    def __getattr__(self, attr):
        if attr == "validator":
            return ADSSeriesValidator(self._feature_type, self._obj)

        for feature_type in self._feature_type:
            if feature_type == ADSString:
                # We need to initialize first to use the plugins.
                if hasattr(feature_type("default"), attr):
                    methods = inspect.classify_class_attrs(feature_type)
                    for method in methods:
                        if method.name == attr:
                            if method.kind == "method":
                                return lambda *args, **kwargs: [
                                    getattr(ADSString(v), attr)(*args, **kwargs)
                                    for v in self._obj.values
                                ]

                            elif method.kind == "property":
                                attr_objects = []
                                for v in self._obj.values:
                                    attr_objects.append(getattr(feature_type(v), attr))
                                return attr_objects

            if hasattr(feature_type, attr):
                # non-instance methods, aka class method or static method
                non_ins_methods = [
                    method.name
                    for method in inspect.classify_class_attrs(feature_type)
                    if method.kind == "class method" or method.kind == "static method"
                ]
                # instance method
                ins_methods = [
                    method.name
                    for method in inspect.classify_class_attrs(feature_type)
                    if method.kind == "method"
                ]
                attr_object = getattr(feature_type, attr)

                # if isinstance(attr_object, FeatureValidator):
                #     attr_object._bind_data(self._obj)

                # there was one case that inspect could not track one of the function.
                # hence replace with __dict__. however, problem with __dict__ is that
                # it does not track back its ancestors functions. hence, use a union
                # to safeguard different scenarios.
                if (attr in non_ins_methods) or (
                    attr in feature_type.__dict__
                    and isinstance(feature_type.__dict__[attr], staticmethod)
                    or isinstance(feature_type.__dict__[attr], classmethod)
                ):
                    return lambda *args, **kwargs: attr_object(
                        self._obj, *args, **kwargs
                    )
                elif attr in ins_methods:
                    return lambda *args, **kwargs: getattr(feature_type(), attr)(
                        self._obj, *args, **kwargs
                    )
                return attr_object

        raise AttributeError(
            f"'{self.__class__.__name__}' does not have an attribute {attr}."
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os

from ads.common.decorator.deprecate import deprecated
from multiprocessing.pool import ThreadPool

try:
    import dask
    from dask.distributed import Client
except ImportError as e:
    raise ModuleNotFoundError("Install dask to use the MLRuntime class.") from e

_default_dask_threadpool_size = 50
_client = None
_storage = None


class MLRuntime(object):
    @deprecated("2.5.2")
    def __init__(self):
        dask_threadpool_size = (
            os.environ.get("DASK_THREADPOOL_SIZE") or _default_dask_threadpool_size
        )
        dask.config.set(pool=ThreadPool(int(dask_threadpool_size)))
        if "DASK_SCHEDULER" in os.environ:
            self.dask_client = Client(address=os.environ["DASK_SCHEDULER"])
        elif os.environ.get("DASK_LOCAL_CLIENT", "False").lower() == "true":
            self.dask_client = Client(
                processes=False,
                local_directory="/tmp",  # use directory with write permits
            )
        else:
            # first let's see if the default client is already running, if it is
            # connect to it, else connect a new Dask scheduler
            # localhost:8786 for dask scheduler
            try:
                self.dask_client = Client("127.0.0.1:8786", timeout=1)
            except:
                self.dask_client = Client(
                    processes=False,  # use threads, avoid for inproc cluster IPC overhead
                    name="Oracle Data Science Local Dask",
                    heartbeat_interval=60000,  # one mimnute heartbeats
                    local_directory="/tmp",  # use directory with write permits
                )

    def get_compute_accelerator(self):
        return self.dask_client


def init():
    global _client

    # ensures the client is created only once
    _client = MLRuntime().get_compute_accelerator()


def compute_accelerator():
    global _client
    if _client is None:
        init()
    return _client



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import importlib

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.dataset import logger


class Recommendation:
    recommendation_types = [
        "constant_column",
        "primary_key",
        "imputation",
        "strong_correlation",
        "positive_class",
        "fix_imbalance",
    ]
    recommendation_type_labels = [
        "Constant Columns",
        "Potential Primary Key Columns",
        "Imputation",
        "Multicollinear Columns",
        "Identify positive label for target",
        "Fix imbalance in dataset",
    ]

    @runtime_dependency(module="ipywidgets", install_from=OptionalDependency.NOTEBOOK)
    def __init__(self, ds, recommendation_transformer):
        self.recommendation_transformer = recommendation_transformer
        self.reco_dict = recommendation_transformer.reco_dict_
        self.orig_reco_dict = copy.deepcopy(self.reco_dict)
        self.ds = ds
        self._clear_output()
        self.fill_nan_dict_values = {}
        self.fill_nan_dict = {}
        self.recommendation_type_index = 0
        self.out = ipywidgets.widgets.Output()

    def _clear_output(self):
        self.boxes = []
        self.column_list = []
        self.message_list = []
        self.action_list = {}
        self.control_buttons = None
        self.info = []

    def _get_on_action_change(self, recommendation_type, column):
        def on_action_change(change):
            for message in self.info:
                message.close()
            if change["type"] == "change" and change["name"] == "value":
                if change["new"] == "Fill missing values with constant":
                    self._show_constant_fill_widget(column)
                else:
                    if change["old"] == "Fill missing values with constant":
                        text = self.fill_nan_dict.pop(column, None)
                        if text is not None:
                            text.close()
                self.reco_dict[recommendation_type][column]["Selected Action"] = change[
                    "new"
                ]

        return on_action_change

    @runtime_dependency(module="ipywidgets", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def _show_constant_fill_widget(self, column):
        if column in self.ds.numeric_columns:
            text = ipywidgets.widgets.FloatText(
                description=column,
                disabled=False,
                style={"description_width": "initial"},
                layout=ipywidgets.Layout(width="auto"),
            )
        else:
            text = ipywidgets.widgets.Text(
                description=column,
                disabled=False,
                style={"description_width": "initial"},
                layout=ipywidgets.Layout(width="auto"),
            )
        # copy previously entered value
        if column in self.fill_nan_dict:
            text.value = self.fill_nan_dict[column].value
        self.fill_nan_dict[column] = text

        from IPython.core.display import display

        display(text)
        if self.control_buttons is not None:
            # self.control_buttons.close()
            self.control_buttons = ipywidgets.HBox(self.control_buttons.children)
            display(self.control_buttons)

    def _get_apply_on_click(self):
        def apply(change):
            self.fill_nan_dict_values = {
                k: v.value for k, v in self.fill_nan_dict.items()
            }
            if self.recommendation_type_index < len(self.recommendation_types):
                self._display()
            else:
                self.recommendation_transformer.reco_dict = self.reco_dict
                self.recommendation_transformer.fill_nan_dict_ = (
                    self.fill_nan_dict_values
                )
                self.ds.recommendation_transformer = self.recommendation_transformer
                with self.out:
                    self.ds.new_ds = self.ds.auto_transform()

        return apply

    def _get_reset_on_click(self):
        def reset(change):
            recommendation_type = self.recommendation_types[
                self.recommendation_type_index
            ]
            self.fill_nan_dict_values = None
            self.reco_dict = copy.deepcopy(self.orig_reco_dict)
            for index, column in enumerate(self.reco_dict[recommendation_type].keys()):
                self.action_list[recommendation_type][index].value = self.reco_dict[
                    recommendation_type
                ][column]["Selected Action"]
            for constant_text in self.fill_nan_dict.values():
                constant_text.close()
            for message in self.info:
                message.close()
            self.ds.new_ds = None

        return reset

    def _repr_html_(self):
        self.show_in_notebook()
        return ""

    def show_in_notebook(self):
        if len(self.reco_dict) == 0:
            logger.info("No recommendations.")
            return

        self._display()

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(module="ipywidgets", install_from=OptionalDependency.NOTEBOOK)
    def _display(self):

        from IPython.core.display import display

        if self.recommendation_type_index != len(self.recommendation_types):
            if (
                self.recommendation_types[self.recommendation_type_index]
                in self.reco_dict
            ):
                recommendation = self.reco_dict[
                    self.recommendation_types[self.recommendation_type_index]
                ]
                if self.recommendation_type_index == 0:
                    for column in recommendation:
                        print(
                            "Column '{0}' is constant and will be dropped".format(
                                column
                            )
                        )
                    self.recommendation_type_index += 1
                    self._display()
                    return
                else:
                    with self.out:
                        display(
                            ipywidgets.HTML(
                                """
                    <font size=4>
                        {}
                    </font>
                    </body""".format(
                                    self.recommendation_type_labels[
                                        self.recommendation_type_index
                                    ]
                                )
                            ),
                            layout=ipywidgets.Layout(display="flex"),
                        )
                        self.action_list[
                            self.recommendation_types[self.recommendation_type_index]
                        ] = []
                        self.column_list = []
                        self.message_list = []
                        self.extra_info_list = []
                        for column in recommendation.keys():
                            messages = ipywidgets.Label(
                                recommendation[column]["Message"],
                                layout=ipywidgets.Layout(
                                    flex="auto", justify_content="flex-start"
                                ),
                            )
                            actions = ipywidgets.widgets.Dropdown(
                                options=recommendation[column]["Action"],
                                value=recommendation[column]["Selected Action"],
                                width="auto",
                            )
                            if (
                                "Up-sample" in recommendation[column]["Action"]
                                and importlib.util.find_spec("imblearn") is None
                            ):
                                extra_info = ipywidgets.HTML(
                                    """Up-sample requires imbalanced-learn python package. <br>
                                            To install run <code> pip install imbalanced-learn </code>.
                                        """,
                                    layout=ipywidgets.Layout(
                                        flex="auto", flex_flow="wrap"
                                    ),
                                )
                                self.extra_info_list.append(extra_info)
                            else:
                                self.extra_info_list.append(
                                    ipywidgets.Label(
                                        "",
                                        layout=ipywidgets.Layout(
                                            flex="auto", justify_content="flex-start"
                                        ),
                                    )
                                )

                            actions.observe(
                                self._get_on_action_change(
                                    self.recommendation_types[
                                        self.recommendation_type_index
                                    ],
                                    column,
                                )
                            )
                            self.column_list.append(
                                ipywidgets.Label(
                                    column
                                    + "(type: {0})".format(
                                        self.ds.sampled_df[column].dtype
                                    ),
                                    layout=ipywidgets.Layout(flex="auto"),
                                    color="grey",
                                )
                            )
                            self.message_list.append(messages)
                            self.action_list[
                                self.recommendation_types[
                                    self.recommendation_type_index
                                ]
                            ].append(actions)
                        display(
                            ipywidgets.HBox(
                                [
                                    ipywidgets.VBox(self.column_list),
                                    ipywidgets.VBox(self.message_list),
                                    ipywidgets.VBox(
                                        self.action_list[
                                            self.recommendation_types[
                                                self.recommendation_type_index
                                            ]
                                        ]
                                    ),
                                    ipywidgets.VBox(self.extra_info_list),
                                ],
                                layout=ipywidgets.Layout(
                                    display="flex",
                                    flex_flow="row",
                                    border="solid 2px",
                                    align_content="stretch",
                                    width="auto",
                                ),
                            )
                        )
            else:
                self.recommendation_type_index += 1
                self._display()
                return
        if self.control_buttons is not None:
            self.control_buttons.close()
            self.control_buttons = None
        else:
            display(self.out)
        self.recommendation_type_index += 1
        for constant_text in self.fill_nan_dict.values():
            constant_text.close()
        for constant_text in self.fill_nan_dict.values():
            self._show_constant_fill_widget(constant_text.description)
        if self.recommendation_type_index < len(self.recommendation_types):
            apply_button = ipywidgets.widgets.Button(
                description="Next", button_style="primary"
            )
        else:
            apply_button = ipywidgets.widgets.Button(
                description="Apply", button_style="primary"
            )
        apply_button.on_click(self._get_apply_on_click())
        reset_button = ipywidgets.widgets.Button(description="Reset All")
        reset_button.on_click(self._get_reset_on_click())
        self.control_buttons = ipywidgets.HBox([apply_button, reset_button])
        with self.out:
            display(self.control_buttons)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np
import pandas as pd
from ads.common import logger
from ads.explanations.base_explainer import LocalExplainer
from ads.explanations.mlx_interface import init_lime_explainer
from ads.common.decorator.deprecate import deprecated


class MLXLocalExplainer(LocalExplainer):
    """
    Local Explainer class.

    Generates explanations for single predictions from machine learning models.
    For tabular and text datasets, supports

        - Binary classification
        - Multi-class classification
        - Regression
    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self):
        super(LocalExplainer, self).__init__()
        self.explainer = None

    def explain(
        self,
        row,
        y=None,
        num_features=None,
        report_fidelity=True,
        validation_percentiles=None,
        selected_features=None,
    ):
        """
        Explains the sample, row. Returns a local explanation object.

        Parameters
        ----------
        row : pandas.DataFrame
            Pandas DataFrame of one row (instance) to explain.
        y : pandas.DataFrame/Series, optional
            True target label/value for row. Default value is `None`.
        num_features : int, optional
            Number of features to show in the local explanation. By default `None`,
            which includes all features.
        report_fidelity : bool, optional
            If `True`, the explanation quality (fidelity) is computed and shown in
            the local explanation visualization. Default value is `True`.
        validation_percentiles : list of int, optional
            List of values specifying the different regions to evaluate the local explanation model
            quality (fidelity). This is specified in percentiles of the distances from the
            instance to explain to all of the samples in the dataset used to fit the explainer.
            For example, [1, 5, 10], generates three evaluation datasets with a maximum distance from
            the instance to explain (row) of 1, 5, and 10 percentiles of the distances from the instance
            to explain (row) and all other samples in the train set. The evaluation dataset at percentile
            1 is very local to the sample to explain, the evaluation dataset at percentile 10 evaluates
            the explanation quality further away from the instance to explain (row). This can be helpful
            to see how the explanation generalizes to nearby samples. By default None ([1, 5, 15]).
        selected_features: list[int], optional
            List of the selected features (list of numbers of features). It can be any subset of
            the original features that are in the dataset provided to the model.
            Default value is None.

        Return
        ------
        LocalExplanation
            Local explanation object.
        """
        self.configure_local_explainer(selected_features=selected_features)
        assert row is not None and isinstance(row, pd.DataFrame) and row.shape[0] == 1

        labels = list(range(len(self.class_names)))

        if num_features is None:
            num_features = len(row.columns.values)

        try:
            explanation = self.explainer.compute(
                row.copy(),
                y=None if y is None else y.copy(),
                labels=labels,
                verbose=False,
                num_features=num_features,
                report_fidelity=report_fidelity,
                validation_percentiles=validation_percentiles,
            )[0]
        except IndexError as e:
            if selected_features is not None:
                raise IndexError(
                    f"Unable to generate local explanations due to: {e}. "
                    f"selected_features must be a list of features within the bounds of the existing features "
                    f"(that were provided to model). Provided selected_features: {selected_features}."
                )
        except Exception as e:
            logger.error(f"Unable to generate local explanations due to: {e}.")
            raise e

        return LocalExplanation(explanation, self.class_names)

    def configure_local_explainer(self, **kwargs):
        """
        Validates the local explanation configuration parameters and initializes the local
        explainer with the provided configuration parameters.

        Supported configuration options in kwargs:
            - `surrogate_model` (str): Surrogate model to use. Can be 'linear' or 'decision_tree'.
            - `num_samples` (int): Number of generated samples to fit the surrogate model.
            - `exp_sorting` (str): Feature importance sorting. Can be 'absolute' or 'ordered'.
            - `scale_weight` (bool): Normalizes the feature importance coefficients from LIME to sum to one.
            - `client`: Only allowed to be None to disable parallelization.
            - `batch_size` (int): Number of local explanations per Dask worker.
            - `random_state` (`None` or `int` or instance of `RandomState`): the random state.
            - `selected_features` (`None`, or `list`): list of the selected features numbers.

        Parameters
        ----------
        **kwargs : dict
            Keyword parameter dictionary.

        Return
        ------
        MLXLocalExplainer
            Modified instance (self)
        """
        avail_args = [
            "client",
            "surrogate_model",
            "num_samples",
            "exp_sorting",
            "scale_weight",
            "batch_size",
            "selected_features",
        ]
        for k, v in kwargs.items():
            if k not in avail_args:
                raise ValueError(
                    "Unexpected argument for the local explainer: {}".format(k)
                )
        if kwargs.get("client", None) is not None:
            raise ValueError(
                "Invalid client provided. Currently only supports disabling parallelization "
                "by setting client=None"
            )
        if kwargs.get("surrogate_model", None) not in ["linear", "decision_tree", None]:
            raise ValueError(
                "Invalid surrogate_model provided. Currently only supports linear or decision_tree"
            )
        selected_features = kwargs.get("selected_features", None)
        if selected_features is not None and not isinstance(selected_features, list):
            raise ValueError(
                f"selected_features ({selected_features}) value must be a list of features, "
                f"but it is of type: {type(selected_features)}."
            )

        self._init_explainer(**kwargs)
        return self

    def summary(self):
        """
        Displays detailed information about the local LIME explainer.

        Return
        ------
        str
            HTML object representing the explainer summary.
        """
        if self.explainer is None:
            self.configure_local_explainer()
        return self.explainer.show_in_notebook()

    def _init_explainer(self, **kwargs):
        """
        Internal function to initialize the local explainer.

        Parameters
        ----------
        **kwargs : dict
            Keyword parameter dictionary.
        """
        if self.mode == "regression":
            self.class_names_ = ["Target"]
        self.explainer = init_lime_explainer(
            self.explainer,
            self.est,
            self.X_train,
            self.y_train,
            self.mode,
            class_names=self.class_names,
            **kwargs,
        )


class LocalExplanation:
    """
    Local explanation object constructed by the :class:`MLXLocalExplainer`.

    Contains functions to visualize the explanation and extract the
    raw explanation data.
    """

    def __init__(self, explanation, class_names):
        self.explanation = explanation
        self.class_names = class_names
        if isinstance(self.class_names, np.ndarray):
            self.class_names = self.class_names.tolist()

    def show_in_notebook(
        self, mode="lime", labels=None, colormap=None, return_wordcloud=False, **kwargs
    ):
        """
        Generate a local explanation visualization for this explanation object.

        Contains:
            - Information about the model's prediction and the true target label/value.
            - The instance being explained.
            - Information about the local explainer configuration parameters.
            - Legend describing how to interpret the explanation.
            - The actual explanation (ordered list of +/- feature importances).
            - Quality (fidelity) evaluation of the explanation.

        Parameters
        ----------
        mode : str
            Type of visualization to generate for the explanation. `mode` can be one of the following:
                - **lime**: Horizontal bar chart where each feature contributes either to an increase
                        or decrease in the target value.
                - **stacked**: Stacked horizontal bar chart where each feature contributes either to an
                        increase or decrease in the target value.
                - **dual**: Horizontal bar chart where each feature can contribute both to an increase
                        or decrease in the target value (only supported when the :class:`MLXLocalExplainer`
                        surrogate model is set to "decision_tree").
                - **wordcloud**: Word cloud highlighting the important features/words for a given target
                             label/value. Features/ or words with higher importance are larger than those
                             of a lower importance.
        labels : tuple, list, int, bool, str
            Label indices or name of label to visualize. If `None`, all of the labels that the
            explanation was generated for will be visualized. Default value is `None`.
        colormap : list of str, optional
            List of standard colormaps to use for the wordclouds. One per label. Defaults to `None`.
        return_wordcloud : bool, optional
            If `True`, the wordcloud objects are returned instead of visualized. Defaults to `False`.
        **kwargs : dict
            Keyword dictionary for configuring the wordclouds.

        Return
        ------
        HTML or pyplot figure
            Returns an HTML object for this local explanation. If `mode=wordcloud` and `return_wordcloud=True`,
            returns the wordcloud
        """
        if labels is None:
            labels = (0,)
        if np.isscalar(labels):
            labels = [labels]
        labels = [
            self.class_names.index(label) if isinstance(label, str) else label
            for label in labels
        ]
        return self.explanation.show_in_notebook(
            mode=mode,
            labels=labels,
            colormap=colormap,
            return_wordcloud=return_wordcloud,
            **kwargs,
        )

    def get_diagnostics(self):
        """
        Extracts the raw explanation and evaluation data from the explanation object to generate
        the visualizations.

        Return
        ------
        dict
            Dictionary containing the raw explanation/evaluation data.
        """
        return self.explanation.get_diagnostic()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import yaml
import os
import click


@click.command()
@click.argument("output_path")
@click.option(
    "--source-folder",
    "-s",
    help="source folder to search for environment yaml",
    required=False,
    default=None,
)
def merge(output_path, source_folder):
    merged = _merge(source_folder)
    with open(output_path, "w") as f:
        yaml.safe_dump(merged, f)


def _merge(source_folder=None):
    conda_dependencies = set([])
    pip_dependencies = set([])
    if not source_folder:
        source_folder = os.path.join("operators")
    for dirpath, dirnames, filenames in os.walk(source_folder):
        for fname in filenames:
            if fname == "environment.yaml":
                env_yaml = os.path.join(dirpath, fname)
                print(env_yaml)
                with open(env_yaml, "r") as f:
                    dependencies = yaml.safe_load(f.read())["dependencies"]
                for dep in dependencies:
                    if isinstance(dep, dict) and "pip" in dep:
                        pip_dependencies.update(dep["pip"])
                    else:
                        conda_dependencies.add(dep)
    conda_dependencies.add("pip")
    merged_dependencies = {
        "dependencies": list(conda_dependencies) + [{"pip": list(pip_dependencies)}]
    }
    return merged_dependencies


if __name__ == "__main__":
    merge()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


# Standard lib
import time

# Third party
from abc import abstractmethod
from tqdm import tqdm_notebook
from oci._vendor import six


class ProgressBar(object):
    """ProgressBar is an abstract class for creating progress bars.

    Methods
    -------
    __enter__()
        runtime context entry method
    update(description)
        abstract method for update
    __exit__(exc_type, exc_val, exc_tb)
        runtime context exit method

    """

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def update(self, description):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TqdmProgressBar(ProgressBar):
    """TqdmProgressBar represents a progress bar for notebooks. It inherits from the
    abstract class ProgressBar

    Attributes
    ----------
    max_progress (int)
        The maximum value for the progress bar
    decription (str)
        The progres bar's description
    progress_bar (tqdm_notebook)
        Notebook widget
    verbose (bool)
        verbosity flag

    Methods
    -------
    __enter__()
        runtime context entry method
    __init__(max_progress, description, verbose)
        init method for class
    update(description=None)
        updates progress bar
    __exit__(exc_type, exc_val, exc_tb)
        runtime context exit method
    """

    def __enter__(self):
        """__enter__ runs when the context starts. It returns the class object"""
        return self

    def __init__(self, max_progress=100, description="Running", verbose=False):
        """Class initialization function

        Args:
            max_progress (int, optional): the maximum progress bar value (defaults to 100)
            description (str, optional): the progress bar description (defaults to "Running")
            verbose (bool, optional): verbosity flag (defaults to False)
        """

        self.max_progress = max_progress
        self.progress_bar = tqdm_notebook(
            range(max_progress), desc="loop1", mininterval=0, leave=False
        )
        self.progress_bar.set_description(description)
        self.verbose = verbose
        if self.verbose:
            self.start_time = time.time()
            self.description = description

    def update(self, description=None):
        """update updates the progress bar

        Args:
            description (str, optional): progress bar description

        Returns:
            Nothing
        """

        if self.verbose and description is not None:
            print(
                "%s:%ss"
                % (
                    self.description,
                    str.format("{0:.3f}", time.time() - self.start_time),
                )
            )
            self.description = (
                description if description is not None else self.description
            )
            self.start_time = time.time()
        self.progress_bar.update(1)
        if description is not None:
            self.progress_bar.set_description(description)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """__exit__ runs when the context closes. On success, i.e. all arguments are None
        it sets the description and closes the context. Otherwise it reraises an exception
        thrown in the context

        Args:
            exc_type (Exception): exception type
            exc_val (Exception value): exception value
            exc_tb (Traceback): exception traceback

        Returns:
            Nothing
        """

        if exc_tb is None:
            self.progress_bar.set_description("Done")
            self.progress_bar.close()
            return
        six.reraise(exc_type, exc_val, exc_tb)


class DummyProgressBar(ProgressBar):
    """DummyProgressBar is represents a progress bar for non-notebook environments. It inherits
    from the abstract class ProgressBar. It allows use of the same contextlib enter and exit methods
    """

    def __enter__(self):
        return self

    def __init__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        return



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""
This is a driver script auto-generated by Oracle ADS to run Python script in OCI Data Science Jobs.

The following environment variables are used:
CODE_ENTRYPOINT:
    Relative path to the entry script/module file in the user code directory.
ENTRY_FUNCTION:
    Optional, function name in the entry script/module to be invoked.
    If this is not specified, the entry script will be run as Python script.
PYTHON_PATH:
    Optional, additional paths to be added to sys.path for looking up modules and packages.
    The root of the user code will be added by default.
    Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac, semicolon(;) for Windows.
OUTPUT_DIR:
    Optional, output directory to be copied to object storage.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
WORKING_DIR:
    Optional, the working directory of the user code.
    This can be specified as a relative path from /home/datascience/decompressed_artifact/code
    If this is not specified, the working directory will be /home/datascience/decompressed_artifact/code

This module requires the following package:
oci


"""
try:
    # This is used by ADS and testing
    from .driver_utils import OCIHelper, JobRunner
except ImportError:
    # This is used when the script is in a job run.
    from driver_utils import OCIHelper, JobRunner


def main():
    """The main function for running the job."""
    JobRunner().set_working_dir().conda_unpack().setup_python_path().run()
    OCIHelper.copy_outputs()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import sys

# TODO - Revisit this as part of ADS logging changes https://jira.oci.oraclecorp.com/browse/ODSC-36245
# Use a unique logger that we can individually configure without impacting other log statements.
# We don't want the logger name to mention "ads", since this logger will report any exception that happens in a
# notebook cell, and we don't want customers incorrectly assuming that ADS is somehow responsible for every error.
logger = logging.getLogger("ipython.traceback")
# Set propagate to False so logs aren't passed back up to the root logger handlers. There are some places in ADS
# where logging.basicConfig() is called. This changes root logger configurations. The user could import/use code that
# invokes the logging.basicConfig() function at any time, making the behavior of the root logger unpredictable.
logger.propagate = False
logger.handlers.clear()
traceback_handler = logging.StreamHandler()
traceback_handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
logger.addHandler(traceback_handler)


def _log_traceback(self, exc_tuple=None, **kwargs):
    try:
        etype, value, tb = self._get_exc_info(exc_tuple)
    except ValueError:
        print("No traceback available to show.", file=sys.stderr)
        return
    msg = etype.__name__, str(value)
    # User a generic message that makes no mention of ADS.
    logger.error("Exception", exc_info=(etype, value, tb))
    sys.stderr.write("{0}: {1}".format(*msg))


def set_ipython_traceback():
    pass


def configure_plotting():
    try:
        import IPython
        from IPython import get_ipython
        from IPython.core.error import UsageError

        ipy = get_ipython()
        if ipy is not None:
            try:
                # show matplotlib plots inline
                ipy.run_line_magic("matplotlib", "inline")
            except UsageError:
                #  ignore error and use the default matplotlib mode
                pass
    except ModuleNotFoundError:
        import matplotlib as mpl

        mpl.rcParams["backend"] = "agg"
        import matplotlib.pyplot as plt

        plt.switch_backend("agg")


try:
    import IPython

    global orig_ipython_traceback
    if IPython.core.interactiveshell.InteractiveShell.showtraceback != _log_traceback:
        orig_ipython_traceback = (
            IPython.core.interactiveshell.InteractiveShell.showtraceback
        )

    # Override the default showtraceback behavior of ipython, to show only the error message and log the stacktrace
    IPython.core.interactiveshell.InteractiveShell.showtraceback = _log_traceback
except ModuleNotFoundError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import os
from xml.dom import minidom
from xml.etree.ElementTree import Element
from xml.etree import ElementTree as et

from ads.common.auth import AuthType, create_signer
from ads.opctl.config.utils import read_from_ini
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.merger import ConfigMerger
from ads.opctl.utils import get_oci_region, is_in_notebook_session
from ads.opctl.constants import DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR


def core_site(auth, oci_config, overwrite, oci_profile):

    if is_in_notebook_session():
        if not os.path.exists(
            os.path.join(DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR, "spark-defaults.conf")
        ):
            raise RuntimeError("Pyspark conda pack is not installed.")
        core_site_loc = os.path.join(
            DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR, "core-site.xml"
        )
    else:
        if "CONDA_PREFIX" not in os.environ or not os.path.exists(
            os.path.join(os.environ["CONDA_PREFIX"], "spark-defaults.conf")
        ):
            raise RuntimeError("Please run this inside a Pyspark conda environment.")
        core_site_loc = os.path.join(os.environ["CONDA_PREFIX"], "core-site.xml")

    if os.path.exists(core_site_loc) and not overwrite:
        print(f"{core_site_loc} already exists. Please use `--overwrite` option.")
        return

    p = ConfigProcessor().step(
        ConfigMerger,
        oci_config=oci_config,
        oci_profile=oci_profile,
        auth=auth,
    )
    exec_config = p.config["execution"]
    oci_config = exec_config["oci_config"]
    oci_profile = exec_config["oci_profile"]
    auth = exec_config["auth"]

    properties = generate_core_site_properties(auth, oci_config, oci_profile)
    xmlstr = generate_core_site_properties_str(properties)

    with open(core_site_loc, "w") as f:
        f.write(xmlstr)

    print(f"The core-site.xml is being written to {core_site_loc}")


def generate_core_site_properties(
    authentication,
    oci_config=None,
    oci_profile=None,
):

    region = os.getenv("NB_REGION", None)
    if oci_config and oci_profile:
        region = region or get_oci_region(
            create_signer(AuthType.API_KEY, oci_config, oci_profile)
        )

    if authentication == "api_key":
        if not oci_config or not oci_profile:
            raise ValueError(
                "oci-config and oci-profile must be provided when using api_key."
            )
        oci_config_path = os.path.abspath(os.path.expanduser(oci_config))

        if not os.path.exists(oci_config_path):
            raise RuntimeError(
                f"The oci config path {oci_config_path} does not exist. Please refer to the getting-started.ipynb on generating oci API keys."
            )

        config = read_from_ini(oci_config_path)
        if oci_profile not in config:
            raise RuntimeError(
                f"The configuration profile {oci_profile} does not exist in {oci_config_path}"
            )
        details_dict = config[oci_profile]

        for key in ["user", "fingerprint", "key_file", "tenancy", "region"]:
            if key not in details_dict.keys():
                raise ValueError(
                    f"{key} not found in oci config. Is {oci_config_path} is corrupt?"
                )

        # setup region
        details_dict["region"] = region

        return [
            (
                "fs.oci.client.hostname",
                f"https://objectstorage.{details_dict['region']}.oraclecloud.com",
            ),
            ("fs.oci.client.auth.tenantId", details_dict["tenancy"]),
            ("fs.oci.client.auth.userId", details_dict["user"]),
            ("fs.oci.client.auth.fingerprint", details_dict["fingerprint"]),
            (
                "fs.oci.client.auth.pemfilepath",
                os.path.expanduser(details_dict["key_file"]),
            ),
        ]

    else:
        return [
            (
                "fs.oci.client.hostname",
                f"https://objectstorage.{region}.oraclecloud.com",
            ),
            (
                "fs.oci.client.custom.authenticator",
                "com.oracle.bmc.hdfs.auth.ResourcePrincipalsCustomAuthenticator",
            ),
        ]


def generate_core_site_properties_str(properties):
    def core_site_property(name, value):
        elem = Element("property")
        child_name = Element("name")
        child_name.text = name
        child_value = Element("value")
        child_value.text = value
        elem.append(child_name)
        elem.append(child_value)
        return elem

    elem = Element("configuration")
    for (name, value) in properties:
        child = core_site_property(name, value)
        elem.append(child)

    return (
        minidom.parseString(et.tostring(elem)).childNodes[0].toprettyxml(indent="   ")
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Text feature type.

Classes:
    Text
        The Text feature type.
"""
import matplotlib.pyplot as plt
import pandas as pd
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import random_color_func, SchemeNeutral

from ads.common import utils, logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class Text(String):
    """
    Type representing text values.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_plot(x: pd.Series) -> plt.Axes
        Shows distributions of datasets using wordcloud.
    """

    @staticmethod
    @runtime_dependency(module="wordcloud", install_from=OptionalDependency.TEXT)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows distributions of datasets using wordcloud.

        Examples
        --------
        >>> text = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='text')
        >>> text.ads.feature_type = ['text']
        >>> text.ads.feature_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Text feature type.
        """
        col_name = x.name if x.name else "text"
        df = x.to_frame(col_name)
        words = df[col_name].dropna().to_list()
        words = " ".join([s for s in words if isinstance(s, str)])
        if not words:
            return

        from wordcloud import WordCloud

        wc = WordCloud(
            background_color=SchemeNeutral.BACKGROUND_LIGHT,
            color_func=random_color_func,
        ).generate(words)
        _, ax = plt.subplots(facecolor=SchemeNeutral.BACKGROUND_LIGHT)
        ax.imshow(wc)
        plt.axis("off")
        return ax

    description = "Type representing text values."

    @classmethod
    def feature_domain(cls):
        """
        Returns
        -------
        None
            Nothing.
        """
        return None



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from typing import Dict
from ads.jobs.builders.infrastructure.dsc_job import DataScienceJob


class CustomScriptStep(DataScienceJob):
    def __init__(self, spec: Dict = None, **kwargs) -> None:
        """Initialize a custom script step infrastructure.

        Example
        -------
        Here is an example for defining a custom script step infrastructure using builder:

        .. code-block:: python

            from ads.pipeline import CustomScriptStep
            # Define an OCI Data Science custom script step infrastructure
            infrastructure = (
                CustomScriptStep()
                .with_shape_name("VM.Standard2.1")
                .with_block_storage_size(50)
            )

        See Also
        --------
        https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/user_guide/pipeline/index.html
        """
        super().__init__(spec, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import os
import json
import shlex
from typing import Dict

import ads
from ads.opctl.backend.base import Backend
from ads.common.auth import create_signer, AuthContext
from ads.common.oci_client import OCIClientFactory

from ads.jobs import Job, DataFlow, DataFlowRuntime, DataFlowRun

REQUIRED_FIELDS = [
    "compartment_id",
    "driver_shape",
    "executor_shape",
    "logs_bucket_uri",
    "script_bucket",
]


class DataFlowBackend(Backend):
    def __init__(self, config: Dict) -> None:
        """
        Initialize a MLJobBackend object given config dictionary.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config
        self.oci_auth = create_signer(
            config["execution"].get("auth"),
            config["execution"].get("oci_config", None),
            config["execution"].get("oci_profile", None),
        )
        self.auth_type = config["execution"].get("auth")
        self.profile = config["execution"].get("oci_profile", None)
        self.client = OCIClientFactory(**self.oci_auth).dataflow

    def apply(self):
        """
        Create DataFlow and DataFlow Run from YAML.
        """
        # TODO add the logic for build dataflow and dataflow run from YAML.
        raise NotImplementedError(f"`apply` hasn't been supported for data flow yet.")

    def run(self) -> None:
        """
        Create DataFlow and DataFlow Run from OCID or cli parameters.
        """
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            if self.config["execution"].get("ocid", None):
                data_flow_id = self.config["execution"]["ocid"]
                run_id = Job.from_dataflow_job(data_flow_id).run().id
            else:
                infra = self.config.get("infrastructure", {})
                if any(k not in infra for k in REQUIRED_FIELDS):
                    missing = [k for k in REQUIRED_FIELDS if k not in infra]
                    raise ValueError(
                        f"Following fields are missing but are required for OCI DataFlow Jobs: {missing}. Please run `ads opctl configure`."
                    )
                rt_spec = {}
                rt_spec["scriptPathURI"] = os.path.join(
                    self.config["execution"]["source_folder"],
                    self.config["execution"]["entrypoint"],
                )
                if "script_bucket" in infra:
                    rt_spec["scriptBucket"] = infra.pop("script_bucket")
                if self.config["execution"].get("command"):
                    rt_spec["args"] = shlex.split(self.config["execution"]["command"])
                if self.config["execution"].get("archive"):
                    rt_spec["archiveUri"] = self.config["execution"]["archive"]
                    rt_spec["archiveBucket"] = infra.pop("archive_bucket", None)
                rt = DataFlowRuntime(rt_spec)
                if "configuration" in infra:
                    infra["configuration"] = json.loads(infra["configuration"])
                df = Job(infrastructure=DataFlow(spec=infra), runtime=rt)
                df.create(overwrite=self.config["execution"].get("overwrite", False))
                job_id = df.id
                run_id = df.run().id
        print("DataFlow App ID:", job_id)
        print("DataFlow Run ID:", run_id)
        return {"job_id": job_id, "run_id": run_id}

    def cancel(self):
        """
        Cancel DataFlow Run from OCID.
        """
        if not self.config["execution"].get("run_id"):
            raise ValueError("Can only cancel a DataFlow run.")
        run_id = self.config["execution"]["run_id"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            DataFlowRun.from_ocid(run_id).delete()

    def delete(self):
        """
        Delete DataFlow or DataFlow Run from OCID.
        """
        if self.config["execution"].get("id"):
            data_flow_id = self.config["execution"]["id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                Job.from_dataflow_job(data_flow_id).delete()
        elif self.config["execution"].get("run_id"):
            run_id = self.config["execution"]["run_id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                DataFlowRun.from_ocid(run_id).delete()

    def watch(self):
        """
        Watch DataFlow Run from OCID.
        """
        run_id = self.config["execution"]["run_id"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            run = DataFlowRun.from_ocid(run_id)
            run.watch()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
import logging
import time
from typing import List, Optional

import oci
import yaml
from ads.common.decorator.runtime_dependency import (
    OptionalDependency,
    runtime_dependency,
)
from ads.common.extended_enum import ExtendedEnumMeta
from ads.common.oci_datascience import OCIDataScienceMixin
from ads.common.oci_logging import ConsolidatedLog, OCILog
from ads.jobs.builders.infrastructure.base import RunInstance
from ads.pipeline.ads_pipeline_step import PipelineStep
from ads.pipeline.visualizer.base import (
    GraphOrientation,
    PipelineVisualizer,
    StepStatus,
)
from ads.pipeline.visualizer.graph_renderer import PipelineGraphRenderer
from ads.pipeline.visualizer.text_renderer import PipelineTextRenderer

PIPELINE_RUN_TERMINAL_STATE = {
    StepStatus.FAILED,
    StepStatus.SUCCEEDED,
    StepStatus.CANCELED,
    StepStatus.DELETED,
    StepStatus.SKIPPED,
}

LOG_INTERVAL = 3
SLEEP_INTERVAL = 3
MAXIMUM_TIMEOUT_SECONDS = 1800
LOG_RECORDS_LIMIT = 100
ALLOWED_OPERATION_KWARGS = [
    "allow_control_chars",
    "retry_strategy",
    "delete_related_job_runs",
    "if_match",
    "opc_request_id",
]
ALLOWED_WAITER_KWARGS = [
    "max_interval_seconds",
    "max_wait_seconds",
    "wait_callback",
    "fetch_func",
]

logger = logging.getLogger(__name__)


class LogType(str, metaclass=ExtendedEnumMeta):
    CUSTOM_LOG = "custom_log"
    SERVICE_LOG = "service_log"


class ShowMode(str, metaclass=ExtendedEnumMeta):
    GRAPH = "graph"
    TEXT = "text"


class StepType(str, metaclass=ExtendedEnumMeta):
    ML_JOB = "ML_JOB"
    CUSTOM_SCRIPT = "CUSTOM_SCRIPT"


class LogNotConfiguredError(Exception):
    pass


class PipelineRun(
    OCIDataScienceMixin, oci.data_science.models.PipelineRun, RunInstance
):
    """
    Attributes
    ----------
    pipeline: Pipeline
        Returns the ADS pipeline object for run instance.
    status: str
        Returns Lifecycle status.
    custom_logging: OCILog
        Returns the OCILog object containing the custom logs from the pipeline.

    Methods
    -------
    create(self) -> PipelineRun
        Creates an OCI pipeline run.
    delete(self, delete_related_job_runs: Optional[bool] = True, max_wait_seconds: Optional[int] = MAXIMUM_TIMEOUT, **kwargs) -> PipelineRun
        Deletes an OCI pipeline run.
    cancel(self, maximum_timeout: int = MAXIMUM_TIMEOUT) -> PipelineRun
        Cancels an OCI pipeline run.
    watch(self, steps: List[str] = None, interval: float = LOG_INTERVAL, log_type: str = LogType.CUSTOM_LOG, *args) -> PipelineRun
        Watches the pipeline run until it finishes.
    list(cls, pipeline_id: str, compartment_id: Optional[str] = None, **kwargs) -> List[PipelineRun]:
        Lists pipeline runs for a given pipeline.
    to_yaml(self) -> str
        Serializes the object into YAML string.
    show(self, mode: str = ShowMode.GRAPH, wait: bool = False, rankdir: str = GraphOrientation.TOP_BOTTOM) -> None
        Renders pipeline run. Can be `text` or `graph` representation.
    to_svg(self, uri: str = None, rankdir: str = GraphOrientation.TOP_BOTTOM, **kwargs)
        Renders pipeline run graph to SVG.
    sync(self) -> None
        Syncs status of Pipeline run.
    """

    _DETAILS_LINK = (
        "https://console.{region}.oraclecloud.com/data-science/pipeline-runs/{id}"
    )

    def __init__(
        self,
        config: dict = None,
        signer: oci.signer.Signer = None,
        client_kwargs: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(config, signer, client_kwargs, **kwargs)
        self._service_logging = None
        self._custom_logging = None
        self._pipeline = None

        self._graphViz = PipelineVisualizer().with_renderer(
            PipelineGraphRenderer(show_status=True)
        )
        self._textViz = PipelineVisualizer().with_renderer(PipelineTextRenderer())

    def sync(self, **kwargs) -> None:
        """Syncs status of the Pipeline Run.

        Returns
        -------
        None
        """
        super().sync(**kwargs)
        self._graphViz.with_pipeline(self.pipeline).with_pipeline_run(self)
        self._textViz.with_pipeline(self.pipeline).with_pipeline_run(self)
        return self

    def show(
        self,
        mode: str = ShowMode.GRAPH,
        wait: bool = False,
        rankdir: str = GraphOrientation.TOP_BOTTOM,
    ) -> None:
        """
        Renders pipeline run. Can be `text` or `graph` representation.

        Parameters
        ----------
        mode: (str, optional). Defaults to `graph`.
            Pipeline run display mode. Allowed values: `graph` or `text`.
        wait: (bool, optional). Default to `False`.
            Whether to wait until the completion of the pipeline run.
        rankdir: (str, optional). Default to `TB`.
            Direction of the rendered graph. Allowed Values: `TB` or `LR`.
            Applicable only for graph mode.

        Returns
        -------
        None
        """
        self.sync()
        renderer = self._graphViz if mode.lower() == ShowMode.GRAPH else self._textViz
        if not wait:
            renderer.render(rankdir=rankdir)
            return
        self._show(renderer, rankdir=rankdir)

    def to_svg(
        self, uri: str = None, rankdir: str = GraphOrientation.TOP_BOTTOM, **kwargs
    ) -> str:
        """
        Renders pipeline run graph to SVG.

        Parameters
        ----------
        uri: (string, optional). Defaults to `None`.
            URI location to save the SVG string.
        rankdir: (str, optional). Default to `TB`.
            Direction of the rendered graph. Allowed Values: `TB` or `LR`.
            Applicable only for graph mode.

        Returns
        -------
        str
            Pipeline run graph in svg format.
        """
        self.sync()
        return self._graphViz.to_svg(uri=uri, rankdir=rankdir, **kwargs)

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def _show(
        self,
        viz,
        rankdir: str = GraphOrientation.TOP_BOTTOM,
        refresh_interval=SLEEP_INTERVAL,
    ):
        """
        Renders pipeline run in text or graph until the completion of the pipeline.

        Parameters
        ----------
        viz: PipelineRenderer
            The `PipelineTextRenderer` or `PipelineGraphRenderer` object.
        rankdir: (str, optional). Default to `TB`.
            Direction of the rendered graph. Allowed Values: `TB` or `LR`.
            Applicable only for graph mode.
        refresh_interval: (int, optional). Defaults to 5.
            Time interval in seconds to refresh pipeline run status.

        Returns
        -------
        None
        """
        from IPython.display import clear_output

        try:
            while self.status not in PIPELINE_RUN_TERMINAL_STATE:
                clear_output(wait=True)
                viz.render(rankdir=rankdir)
                time.sleep(refresh_interval)

            clear_output(wait=True)
            viz.render(rankdir=rankdir)
        except KeyboardInterrupt:
            pass

    def logs(self, log_type: str = None) -> ConsolidatedLog:
        """Builds the consolidated log for pipeline run.

        Parameters
        ----------
        log_type: str
            The log type of the pipeline run. Defaults to None.
            Can be custom_log, service_log or None.

        Returns
        -------
        ConsolidatedLog
            The ConsolidatedLog instance.
        """
        logging_list = []
        if not log_type:
            try:
                logging_list.append(self.custom_logging)
            except LogNotConfiguredError:
                pass

            try:
                logging_list.append(self.service_logging)
            except LogNotConfiguredError:
                pass

            if not logging_list:
                raise LogNotConfiguredError(
                    "Neither `custom` nor `service` log was configured for the pipeline run."
                )
        elif log_type == LogType.SERVICE_LOG:
            logging_list = [self.service_logging]
        elif log_type == LogType.CUSTOM_LOG:
            logging_list = [self.custom_logging]
        else:
            raise ValueError(
                "Parameter log_type should be either custom_log, service_log or None."
            )

        return ConsolidatedLog(*logging_list)

    @property
    def pipeline(self):
        """Returns the ADS Pipeline instance.
        Step details will be synched with the Pipeline Run.

        Parameters
        ----------
        None

        Returns
        -------
        Pipeline
            The ADS Pipeline instance, where Step details will be synched with the Pipeline Run.
        """
        from ads.pipeline.ads_pipeline import Pipeline

        if not self._pipeline:
            self._pipeline = Pipeline.from_ocid(self.pipeline_id)
            self._sync_step_details()
        return self._pipeline

    @property
    def status(self) -> str:
        """Lifecycle status.

        Returns
        -------
        str
            Status in a string.
        """
        self.sync()
        return self.lifecycle_state

    @property
    def custom_logging(self) -> OCILog:
        """The OCILog object containing the custom logs from the pipeline run."""
        if not self._custom_logging:
            self._check_log_details()

            while not self._stop_condition():
                # Break if pipeline run has log ID.
                if self.log_details.log_id:
                    break
                time.sleep(LOG_INTERVAL)

            self._custom_logging = OCILog(
                id=self.log_details.log_id,
                log_group_id=self.log_details.log_group_id,
                compartment_id=self.compartment_id,
                annotation="custom",
            )
        return self._custom_logging

    @property
    def service_logging(self) -> OCILog:
        """The OCILog object containing the service logs from the pipeline run."""
        if not self._service_logging:
            self._check_log_details()
            self._service_logging = self._get_service_logging()
        return self._service_logging

    def _check_log_details(self):
        if not self.log_details:
            raise LogNotConfiguredError(
                "Pipeline log is not configured. Make sure log group id is added."
            )
        if not self.log_details.log_group_id:
            raise LogNotConfiguredError(
                "Log group OCID is not specified for this pipeline. Call with_log_group_id to add it."
            )

    def _get_service_logging(self) -> OCILog:
        """Builds the OCI service log instance for pipeline run.

        Returns
        -------
        OCILog
            The OCILog instance.
        """
        log_summary = self._search_service_logs()

        if not log_summary:
            raise LogNotConfiguredError("Service log is not configured for pipeline.")

        # each pipeline can only have one service log
        service_log_id = log_summary[0].id
        return OCILog(
            id=service_log_id,
            log_group_id=self.log_details.log_group_id,
            compartment_id=self.compartment_id,
            annotation="service",
        )

    def _search_service_logs(self) -> List[oci.logging.models.log_summary.LogSummary]:
        """Search the service log of pipeline run based on
        log_group_id, source_service, source_resource and log_type.

        Returns
        -------
        list
            A list of oci.logging.models.log_summary.LogSummary.
        """
        return (
            OCILog(compartment_id=self.compartment_id)
            .client.list_logs(
                log_group_id=self.log_details.log_group_id,
                source_service=self.pipeline.CONST_SERVICE,
                source_resource=self.pipeline.id,
                log_type="SERVICE",
            )
            .data
        )

    def _sync_step_details(self) -> None:
        """Combines pipeline step details with override step details.

        Returns
        -------
        None
        """
        if not self._pipeline or not self._pipeline.step_details:
            return None

        updated_step_details = []
        for step in self._pipeline.step_details:
            updated_step_detail = copy.deepcopy(step.to_dict())
            # restore dependencies information
            updated_step_detail["spec"]["dependsOn"] = step.depends_on
            # override step details if necessary
            if self.step_override_details:
                for override_step in self.step_override_details:
                    if type(override_step) == dict:
                        break
                    if step.name == override_step.step_name:
                        if override_step.step_configuration_details:
                            if (
                                "stepConfigurationDetails"
                                not in updated_step_detail["spec"]
                            ):
                                updated_step_detail["spec"][
                                    "stepConfigurationDetails"
                                ] = {}

                            if (
                                override_step.step_configuration_details.maximum_runtime_in_minutes
                            ):
                                updated_step_detail["spec"]["stepConfigurationDetails"][
                                    "maximumRuntimeInMinutes"
                                ] = (
                                    override_step.step_configuration_details.maximum_runtime_in_minutes
                                )
                            if (
                                override_step.step_configuration_details.environment_variables
                            ):
                                updated_step_detail["spec"]["stepConfigurationDetails"][
                                    "environmentVariables"
                                ] = (
                                    override_step.step_configuration_details.environment_variables
                                )
                            if (
                                override_step.step_configuration_details.command_line_arguments
                            ):
                                updated_step_detail["spec"]["stepConfigurationDetails"][
                                    "commandLineArguments"
                                ] = (
                                    override_step.step_configuration_details.command_line_arguments
                                )

            updated_step_details.append(PipelineStep.from_dict(updated_step_detail))
        self._pipeline.with_step_details(updated_step_details)

    def _set_service_logging_resource(self, service_logging: OCILog):
        """Sets the service logging resource for pipeline run.

        Parameters
        ----------
        service_logging: OCILog instance.
            The OCILog instance.
        """
        self._service_logging = service_logging

    def create(self) -> "PipelineRun":
        """Creates an OCI pipeline run.

        Returns
        -------
        PipelineRun:
            Pipeline run instance (self).
        """
        self.load_properties_from_env()
        response = self.client.create_pipeline_run(
            self.to_oci_model(oci.data_science.models.CreatePipelineRunDetails)
        )
        self.update_from_oci_model(response.data)
        return self

    def delete(
        self,
        delete_related_job_runs: Optional[bool] = True,
        max_wait_seconds: Optional[int] = MAXIMUM_TIMEOUT_SECONDS,
        **kwargs,
    ) -> "PipelineRun":
        """Deletes an OCI pipeline run.

        Parameters
        ----------
        delete_related_job_runs: bool, optional
            Specify whether to delete related JobRuns or not. Defaults to True.
        max_wait_seconds: int, optional
            The maximum time to wait, in seconds. Defaults to 1800.

        kwargs: optional
        The kwargs to be executed when deleting the pipeline.
        The allowed keys are:
        * "allow_control_chars": bool, to indicate whether or not this request should
        allow control characters in the response object. By default, the response will
        not allow control characters in strings.
        * "retry_strategy": obj, to apply to this specific operation/call. This will
        override any retry strategy set at the client-level. This should be one of the
        strategies available in the :py:mod:`~oci.retry` module. This operation will not
        retry by default, users can also use the convenient :py:data:`~oci.retry.DEFAULT_RETRY_STRATEGY`
        provided by the SDK to enable retries for it. The specifics of the default retry
        strategy are described `here <https://docs.oracle.com/en-us/iaas/tools/python/latest/sdk_behaviors/retries.html>`__.
        To have this operation explicitly not perform any retries, pass an instance of :py:class:`~oci.retry.NoneRetryStrategy`.
        * "if_match": str, for optimistic concurrency control. In the PUT or DELETE call
        for a resource, set the `if-match` parameter to the value of the etag from a
        previous GET or POST response for that resource. The resource is updated or
        deleted only if the `etag` you provide matches the resource's current `etag` value.
        * "opc_request_id": str, unique Oracle assigned identifier for the request.
        If you need to contact Oracle about a particular request, then provide the request ID.
        * "max_interval_seconds": int, the maximum interval between queries, in seconds.
        * "wait_callback": A function which will be called each time that we have to do an initial
        wait (i.e. because the property of the resource was not in the correct state,
        or the ``evaluate_response`` function returned False). This function should take two
        arguments - the first argument is the number of times we have checked the resource,
        and the second argument is the result of the most recent check.
        * "fetch_func": A function to be called to fetch the updated state from the server.
        This can be used if the call to check for state needs to be more complex than a single
        GET request. For example, if the goal is to wait until an item appears in a list,
        fetch_func can be a function that paginates through a full list on the server.

        Returns
        -------
        PipelineRun:
            Pipeline run instance (self).
        """
        operation_kwargs = {"delete_related_job_runs": delete_related_job_runs}
        waiter_kwargs = {"max_wait_seconds": max_wait_seconds}
        for key, value in kwargs.items():
            if key in ALLOWED_OPERATION_KWARGS:
                operation_kwargs[key] = value
            elif key in ALLOWED_WAITER_KWARGS:
                waiter_kwargs[key] = value

        self.client_composite.delete_pipeline_run_and_wait_for_state(
            pipeline_run_id=self.id,
            wait_for_states=[PipelineRun.LIFECYCLE_STATE_DELETED],
            operation_kwargs=operation_kwargs,
            waiter_kwargs=waiter_kwargs,
        )
        return self.sync()

    def cancel(self, maximum_timeout: int = MAXIMUM_TIMEOUT_SECONDS) -> "PipelineRun":
        """Cancels an OCI pipeline run.

        Parameters
        ----------
        maximum_timeout: int, optional
            The maximum timeout to cancel the pipeline run. Defaults to 1800 seconds.

        Returns
        -------
        PipelineRun:
            Pipeline run instance (self).
        """
        self.client.cancel_pipeline_run(self.id)

        time_counter = 0
        while self.sync().lifecycle_state not in [
            PipelineRun.LIFECYCLE_STATE_CANCELED,
            PipelineRun.LIFECYCLE_STATE_FAILED,
        ]:
            time.sleep(SLEEP_INTERVAL)
            if time_counter > maximum_timeout:
                print(
                    "Pipeline run stopping after ",
                    maximum_timeout,
                    " seconds of not reaching CANCELLED state.",
                )
                break
            time_counter += SLEEP_INTERVAL

        if self.sync().lifecycle_state != PipelineRun.LIFECYCLE_STATE_CANCELED:
            raise Exception("Error occurred in attempt to cancel the pipeline run.")
        return self

    def watch(
        self,
        steps: List[str] = None,
        interval: float = LOG_INTERVAL,
        log_type: str = None,
        *args,
    ) -> "PipelineRun":
        """Watches the pipeline run until it finishes.
        This method will keep streamming the service log of the pipeline run until it's succeeded, failed or cancelled.

        Parameters
        ----------
        steps: (List[str], optional). Defaults to None.
            Pipeline steps passed in to filter the logs.
        interval: (float, optional). Defaults to 3 seconds.
            Time interval in seconds between each request to update the logs.
        log_type: (str, optional). Defaults to None.
            The log type. Can be `custom_log`, `service_log` or None.
        *args:
            Pipeline steps passed in to filter the logs.
            Example: `.watch("step1", "step2")`

        Examples
        --------
        >>> .watch()
        >>> .watch(log_type="service_log")
        >>> .watch("step1", "step2", log_type="custom_log", interval=3)
        >>> .watch(steps=["step1", "step2"], log_type="custom_log", interval=3)

        Returns
        -------
        PipelineRun:
            Pipeline run instance (self).
        """
        logging = self.logs(log_type=log_type)

        steps_to_monitor = list(set(steps or ()) | set(args))

        try:
            return self.__stream_log(
                logging,
                steps_to_monitor,
                interval,
                log_type,
            )
        except KeyboardInterrupt:
            print("Stop watching logs.")
            pass

    def __stream_log(
        self,
        logging: ConsolidatedLog,
        pipeline_steps: List = None,
        interval: float = LOG_INTERVAL,
        log_type: str = None,
    ) -> "PipelineRun":
        """Stream logs from OCI pipeline backends.

        Parameters
        ----------
        logging : ConsolidatedLog.
            The ConsolidatedLog instance.
        pipeline_steps: list
            A list of pipeline step name.
        interval : float
            Time interval in seconds between each request to update the logs.
        log_type : str
            The log type.

        Returns
        -------
        PipelineRun:
            Pipeline run instance (self).
        """
        print(f"Pipeline OCID: {self.pipeline_id}")
        print(f"Pipeline Run OCID: {self.id}")

        if self.time_accepted:
            count = logging.stream(
                interval=interval,
                stop_condition=self._stop_condition,
                time_start=self.time_accepted,
                log_filter=self._build_filter_expression(pipeline_steps, log_type),
            )
            if not count:
                print(
                    "No logs in the last 14 days. Please set time_start to see older logs."
                )

        return self.sync()

    def _build_filter_expression(self, steps: List = [], log_type: str = None) -> str:
        """Builds query expression for logs that are generated by pipeline run and job run.
            The query expression consists of two parts:
            1. Logs that are generated by pipeline run:
                - service and custom logs for CUSTOM_SCRIPT step
                - service log for ML_JOB step
                Format: (source = *<pipeline_run_id> AND ( subject = <pipeline_step_name> OR subject = <pipeline_step_name> OR ...))
            2. Logs that are generated by job run:
                - custom log for ML_JOB step
                Format: source = *<job_run_id> OR source = *<job_run_id> OR source = *<job_run_id> OR ...

            TODO:
                This is a temporary solution, and the real fix will be done after the jobs service add pipleine run details in the log data panel.

        Parameters
        ----------
        steps: list
            A list of pipeline step name.
        log_type : str
            The log type.

        Returns
        -------
        str:
            Query string to search the logs of pipeline.
        """
        sources = []
        subjects = []
        skipped_step_list = []
        for step_run in self.step_runs:
            if not steps or (step_run.step_name in steps):
                step_name = step_run.step_name
                if step_run.step_type == StepType.ML_JOB:
                    if not step_run.job_run_id:
                        skipped_step_list.append(step_run.step_name)
                        continue
                    job_run_id = step_run.job_run_id
                    if log_type == LogType.CUSTOM_LOG:
                        sources.append(f"source = '*{job_run_id}'")
                    elif log_type == LogType.SERVICE_LOG:
                        subjects.append(f"subject = '{step_name}'")
                    else:
                        sources.append(f"source = '*{job_run_id}'")
                        subjects.append(f"subject = '{step_name}'")
                else:
                    subjects.append(f"subject = '{step_name}'")

        if skipped_step_list:
            logger.warning(
                f"ML Jobs: {', '.join(skipped_step_list)} log can't be printed since their job run ids are not known at this time. Please stop and rerun the watch() command again to retrieve the job run ids and print the logs."
            )

        filter_list = []

        # add query for logs that are generated by pipeline run
        if subjects:
            pipeline_log_filters = [f"source = '*{self.id}'"]
            pipeline_log_filters.append("(" + " OR ".join(subjects) + ")")
            filter_list = ["(" + " AND ".join(pipeline_log_filters) + ")"]

        # add query for logs that are generated by job run
        if sources:
            filter_list.extend(sources)

        return " OR ".join(filter_list)

    def _stop_condition(self):
        """Stops the sync once the job is in a terminal state."""
        self.sync()
        return self.lifecycle_state in PIPELINE_RUN_TERMINAL_STATE

    @classmethod
    def list(
        cls, pipeline_id: str, compartment_id: Optional[str] = None, **kwargs
    ) -> List["PipelineRun"]:
        """
        List pipeline runs for a given pipeline.

        Parameters
        ----------
        pipeline_id: str.
            The OCID of pipeline.
        compartment_id: (str, optional). Defaults to None.
            The OCID of compartment.
            If `None`, the value will be taken from the environment variables.
        kwargs
            Additional keyword arguments for filtering pipelines.
            - lifecycle_state: str. Allowed values: "CREATING", "ACTIVE", "DELETING", "FAILED", "DELETED"
            - created_by: str
            - limit: int

        Returns
        -------
        List[PipelineRun]
            The list of pipeline runs.
        """
        PipelineRun.list_resource(compartment_id, pipeline_id=pipeline_id, **kwargs)

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())
